#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import atexit
import contextlib
import functools
import inspect
import json
import logging
import os
import re
import subprocess
import sys
import tempfile
import time
import unittest

import edgedb


log = logging.getLogger(__name__)


@contextlib.contextmanager
def silence_asyncio_long_exec_warning():
    def flt(log_record):
        msg = log_record.getMessage()
        return not msg.startswith('Executing ')

    logger = logging.getLogger('asyncio')
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


def _get_wsl_path(win_path):
    return (
        re.sub(r'^([A-Z]):', lambda m: f'/mnt/{m.group(1)}', win_path)
        .replace("\\", '/')
        .lower()
    )


_default_cluster = None


def _start_cluster(*, cleanup_atexit=True):
    global _default_cluster

    if isinstance(_default_cluster, Exception):
        # when starting a server fails
        # don't retry starting one for every TestCase
        # because repeating the failure can take a while
        raise _default_cluster

    if _default_cluster:
        return _default_cluster

    try:
        tmpdir = tempfile.TemporaryDirectory()
        status_file = os.path.join(tmpdir.name, 'server-status')

        # if running on windows adjust the path for WSL
        status_file_unix = _get_wsl_path(status_file)

        env = os.environ.copy()
        # Make sure the PYTHONPATH of _this_ process does
        # not interfere with the server's.
        env.pop('PYTHONPATH', None)

        edgedb_server = env.get('EDGEDB_SERVER_BINARY', 'edgedb-server')
        args = [
            edgedb_server,
            "--temp-dir",
            "--testmode",
            f"--emit-server-status={status_file_unix}",
            "--port=auto",
            "--auto-shutdown",
            "--bootstrap-command=ALTER ROLE edgedb { SET password := 'test' }",
        ]

        help_args = [edgedb_server, "--help"]
        if sys.platform == 'win32':
            help_args = ['wsl', '-u', 'edgedb'] + help_args

        if "--generate-self-signed-cert" in subprocess.run(
            help_args,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
            env=env,
            cwd=tmpdir.name,
        ).stdout:
            args.append("--generate-self-signed-cert")

        if sys.platform == 'win32':
            args = ['wsl', '-u', 'edgedb'] + args

        if env.get('EDGEDB_DEBUG_SERVER'):
            server_stdout = None
        else:
            server_stdout = subprocess.DEVNULL

        p = subprocess.Popen(
            args,
            env=env,
            cwd=tmpdir.name,
            stdout=server_stdout,
            stderr=subprocess.STDOUT,
        )

        for _ in range(250):
            try:
                with open(status_file, 'rb') as f:
                    for line in f:
                        if line.startswith(b'READY='):
                            break
                    else:
                        raise RuntimeError('not ready')
                break
            except Exception:
                time.sleep(1)
        else:
            raise RuntimeError('server status file not found')

        data = json.loads(line.split(b'READY=')[1])
        con_args = dict(host='localhost', port=data['port'])
        if 'tls_cert_file' in data:
            if sys.platform == 'win32':
                con_args['tls_ca_file'] = os.path.join(
                    tmpdir.name, "edbtlscert.pem"
                )
                subprocess.check_call(
                    [
                        'wsl',
                        '-u',
                        'edgedb',
                        'cp',
                        data['tls_cert_file'],
                        _get_wsl_path(con_args['tls_ca_file'])
                    ]
                )
            else:
                con_args['tls_ca_file'] = data['tls_cert_file']

        con = edgedb.connect(password='test', **con_args)
        _default_cluster = {
            'proc': p,
            'con': con,
            'con_args': con_args,
        }

        if 'tls_cert_file' in data:
            # Keep the temp dir which we also copied the cert from WSL
            _default_cluster['_tmpdir'] = tmpdir

        atexit.register(con.close)
    except Exception as e:
        _default_cluster = e
        raise e

    return _default_cluster


class TestCaseMeta(type(unittest.TestCase)):
    _database_names = set()

    @staticmethod
    def _iter_methods(bases, ns):
        for base in bases:
            for methname in dir(base):
                if not methname.startswith('test_'):
                    continue

                meth = getattr(base, methname)
                if not inspect.iscoroutinefunction(meth):
                    continue

                yield methname, meth

        for methname, meth in ns.items():
            if not methname.startswith('test_'):
                continue

            if not inspect.iscoroutinefunction(meth):
                continue

            yield methname, meth

    @classmethod
    def wrap(mcls, meth):
        @functools.wraps(meth)
        def wrapper(self, *args, __meth__=meth, **kwargs):
            try_no = 1

            while True:
                try:
                    # There might be unobvious serializability
                    # anomalies across the test suite, so, rather
                    # than hunting them down every time, simply
                    # retry the test.
                    self.loop.run_until_complete(
                        __meth__(self, *args, **kwargs))
                except edgedb.TransactionSerializationError:
                    if try_no == 3:
                        raise
                    else:
                        self.loop.run_until_complete(self.con.execute(
                            'ROLLBACK;'
                        ))
                        try_no += 1
                else:
                    break

        return wrapper

    @classmethod
    def add_method(mcls, methname, ns, meth):
        ns[methname] = mcls.wrap(meth)

    def __new__(mcls, name, bases, ns):
        for methname, meth in mcls._iter_methods(bases, ns.copy()):
            if methname in ns:
                del ns[methname]
            mcls.add_method(methname, ns, meth)

        cls = super().__new__(mcls, name, bases, ns)
        if not ns.get('BASE_TEST_CLASS') and hasattr(cls, 'get_database_name'):
            dbname = cls.get_database_name()

            if name in mcls._database_names:
                raise TypeError(
                    f'{name} wants duplicate database name: {dbname}')

            mcls._database_names.add(name)

        return cls


class TestCase(unittest.TestCase, metaclass=TestCaseMeta):
    @classmethod
    def setUpClass(cls):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        cls.loop = loop

    @classmethod
    def tearDownClass(cls):
        cls.loop.close()
        asyncio.set_event_loop(None)

    def add_fail_notes(self, **kwargs):
        if not hasattr(self, 'fail_notes'):
            self.fail_notes = {}
        self.fail_notes.update(kwargs)

    @contextlib.contextmanager
    def annotate(self, **kwargs):
        # Annotate the test in case the nested block of code fails.
        try:
            yield
        except Exception:
            self.add_fail_notes(**kwargs)
            raise

    @contextlib.contextmanager
    def assertRaisesRegex(self, exception, regex, msg=None,
                          **kwargs):
        with super().assertRaisesRegex(exception, regex, msg=msg):
            try:
                yield
            except BaseException as e:
                if isinstance(e, exception):
                    for attr_name, expected_val in kwargs.items():
                        val = getattr(e, attr_name)
                        if val != expected_val:
                            raise self.failureException(
                                f'{exception.__name__} context attribute '
                                f'{attr_name!r} is {val} (expected '
                                f'{expected_val!r})') from e
                raise

    def addCleanup(self, func, *args, **kwargs):
        @functools.wraps(func)
        def cleanup():
            res = func(*args, **kwargs)
            if inspect.isawaitable(res):
                self.loop.run_until_complete(res)
        super().addCleanup(cleanup)


class ClusterTestCase(TestCase):

    BASE_TEST_CLASS = True

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.cluster = _start_cluster(cleanup_atexit=True)


class ConnectedTestCaseMixin:

    @classmethod
    async def connect(cls, *,
                      cluster=None,
                      database='edgedb',
                      user='edgedb',
                      password='test'):
        conargs = cls.get_connect_args(
            cluster=cluster, database=database, user=user, password=password)
        return await edgedb.async_connect_raw(**conargs)

    @classmethod
    def get_connect_args(cls, *,
                         cluster=None,
                         database='edgedb',
                         user='edgedb',
                         password='test'):
        conargs = cls.cluster['con_args'].copy()
        conargs.update(dict(user=user,
                            password=password,
                            database=database))
        return conargs


class DatabaseTestCase(ClusterTestCase, ConnectedTestCaseMixin):
    SETUP = None
    TEARDOWN = None
    SCHEMA = None

    SETUP_METHOD = None
    TEARDOWN_METHOD = None

    # Some tests may want to manage transactions manually,
    # in which case ISOLATED_METHODS will be False.
    ISOLATED_METHODS = True
    # Turns on "EdgeDB developer" mode which allows using restricted
    # syntax like FROM SQL and similar. It allows modifying standard
    # library (e.g. declaring casts).
    INTERNAL_TESTMODE = True

    BASE_TEST_CLASS = True

    def setUp(self):
        if self.INTERNAL_TESTMODE:
            self.loop.run_until_complete(
                self.con.execute(
                    'CONFIGURE SESSION SET __internal_testmode := true;'))

        if self.ISOLATED_METHODS:
            self.loop.run_until_complete(
                self.con.execute(
                    "CONFIGURE SESSION SET session_idle_transaction_timeout "
                    ":= <std::duration>'3m'"
                )
            )
            self.xact = self.con.raw_transaction()
            self.loop.run_until_complete(self.xact.start())

        if self.SETUP_METHOD:
            self.loop.run_until_complete(
                self.con.execute(self.SETUP_METHOD))

        super().setUp()

    def tearDown(self):
        try:
            if self.TEARDOWN_METHOD:
                self.loop.run_until_complete(
                    self.con.execute(self.TEARDOWN_METHOD))
        finally:
            try:
                if self.ISOLATED_METHODS:
                    self.loop.run_until_complete(self.xact.rollback())
                    del self.xact

                if self.con.is_in_transaction():
                    self.loop.run_until_complete(
                        self.con.execute('ROLLBACK'))
                    raise AssertionError(
                        'test connection is still in transaction '
                        '*after* the test')

                if not self.ISOLATED_METHODS:
                    self.loop.run_until_complete(
                        self.con.execute('RESET ALIAS *;'))

            finally:
                super().tearDown()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dbname = cls.get_database_name()

        cls.admin_conn = None
        cls.con = None

        class_set_up = os.environ.get('EDGEDB_TEST_CASES_SET_UP')

        # Only open an extra admin connection if necessary.
        if not class_set_up:
            script = f'CREATE DATABASE {dbname};'
            cls.admin_conn = cls.loop.run_until_complete(cls.connect())
            cls.loop.run_until_complete(cls.admin_conn.execute(script))

        cls.con = cls.loop.run_until_complete(cls.connect(database=dbname))

        if not class_set_up:
            script = cls.get_setup_script()
            if script:
                # The setup is expected to contain a CREATE MIGRATION,
                # which needs to be wrapped in a transaction.
                tx = cls.con.raw_transaction()
                cls.loop.run_until_complete(tx.start())
                cls.loop.run_until_complete(tx.execute(script))
                cls.loop.run_until_complete(tx.commit())
                del tx

    @classmethod
    def get_database_name(cls):
        if cls.__name__.startswith('TestEdgeQL'):
            dbname = cls.__name__[len('TestEdgeQL'):]
        elif cls.__name__.startswith('Test'):
            dbname = cls.__name__[len('Test'):]
        else:
            dbname = cls.__name__

        return dbname.lower()

    @classmethod
    def get_setup_script(cls):
        script = ''

        # Look at all SCHEMA entries and potentially create multiple
        # modules, but always create the 'test' module.
        schema = ['\nmodule test {}']
        for name, val in cls.__dict__.items():
            m = re.match(r'^SCHEMA(?:_(\w+))?', name)
            if m:
                module_name = (m.group(1) or 'test').lower().replace(
                    '__', '.')

                with open(val, 'r') as sf:
                    module = sf.read()

                schema.append(f'\nmodule {module_name} {{ {module} }}')

        # Don't wrap the script into a transaction here, so that
        # potentially it's easier to stitch multiple such scripts
        # together in a fashion similar to what `edb inittestdb` does.
        script += f'\nSTART MIGRATION TO {{ {"".join(schema)} }};'
        script += f'\nPOPULATE MIGRATION; \nCOMMIT MIGRATION;'

        if cls.SETUP:
            if not isinstance(cls.SETUP, (list, tuple)):
                scripts = [cls.SETUP]
            else:
                scripts = cls.SETUP

            for scr in scripts:
                if '\n' not in scr and os.path.exists(scr):
                    with open(scr, 'rt') as f:
                        setup = f.read()
                else:
                    setup = scr

                script += '\n' + setup

        return script.strip(' \n')

    @classmethod
    def tearDownClass(cls):
        script = ''

        class_set_up = os.environ.get('EDGEDB_TEST_CASES_SET_UP')

        if cls.TEARDOWN and not class_set_up:
            script = cls.TEARDOWN.strip()

        try:
            if script:
                cls.loop.run_until_complete(
                    cls.con.execute(script))
        finally:
            try:
                cls.loop.run_until_complete(cls.con.aclose())

                if not class_set_up:
                    dbname = cls.get_database_name()
                    script = f'DROP DATABASE {dbname};'

                    cls.loop.run_until_complete(
                        cls.admin_conn.execute(script))

            except Exception:
                log.exception('error running teardown')
                # skip the exception so that original error is shown instead
                # of finalizer error
            finally:
                try:
                    if cls.admin_conn is not None:
                        cls.loop.run_until_complete(
                            cls.admin_conn.aclose())
                finally:
                    super().tearDownClass()


class AsyncQueryTestCase(DatabaseTestCase):
    BASE_TEST_CLASS = True


class SyncQueryTestCase(DatabaseTestCase):
    BASE_TEST_CLASS = True

    def setUp(self):
        super().setUp()

        cls = type(self)
        cls.async_con = cls.con

        conargs = cls.get_connect_args().copy()
        conargs.update(dict(database=cls.async_con.dbname))

        cls.con = edgedb.connect(**conargs)

    def tearDown(self):
        cls = type(self)
        cls.con.close()
        cls.con = cls.async_con
        del cls.async_con


_lock_cnt = 0


def gen_lock_key():
    global _lock_cnt
    _lock_cnt += 1
    return os.getpid() * 1000 + _lock_cnt


if os.environ.get('USE_UVLOOP'):
    import uvloop
    uvloop.install()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio

from edgedb import errors
from edgedb.pgproto.pgproto cimport (
    WriteBuffer,
    ReadBuffer,
)

from . cimport protocol


cdef class AsyncIOProtocol(protocol.SansIOProtocol):

    def __init__(self, con_params, loop, tls_compat=False):
        protocol.SansIOProtocol.__init__(self, con_params, tls_compat)

        self.loop = loop
        self.transport = None
        self.connected_fut = loop.create_future()
        self.disconnected_fut = None

        self.msg_waiter = None

    cpdef abort(self):
        self.connected = False
        self.terminate()
        if self.transport is not None:
            self.transport.close()
            self.transport = None

    cdef write(self, WriteBuffer buf):
        if self.transport is None:
            raise errors.ClientConnectionFailedTemporarilyError()
        self.transport.write(memoryview(buf))

    async def wait_for_message(self):
        if self.buffer.take_message():
            return

        try:
            self.msg_waiter = self.loop.create_future()
            await self.msg_waiter
            return
        except asyncio.CancelledError:
            # TODO: A proper cancellation requires server/protocol
            # support, which isn't yet available.  Therefore,
            # we're disabling asyncio cancellation completely
            # until we can implement it properly.
            try:
                self.cancelled = True
                self.abort()
            finally:
                raise

    async def try_recv_eagerly(self):
        pass

    async def wait_for_connect(self):
        if self.connected_fut is not None:
            await self.connected_fut

    async def wait_for_disconnect(self):
        if not self.connected:
            return
        else:
            self.disconnected_fut = self.loop.create_future()
            try:
                await self.disconnected_fut
            except ConnectionError:
                pass
            finally:
                self.disconnected_fut = None

    def connection_made(self, transport):
        if self.transport is not None:
            raise RuntimeError('connection_made: invalid connection status')
        self.transport = transport
        self.connected_fut.set_result(True)
        self.connected_fut = None

    def connection_lost(self, exc):
        self.connected = False

        if self.connected_fut is not None and not self.connected_fut.done():
            self.connected_fut.set_exception(ConnectionAbortedError())

        if (
            self.disconnected_fut is not None
            and not self.disconnected_fut.done()
        ):
            self.disconnected_fut.set_exception(ConnectionResetError())

        if self.msg_waiter is not None and not self.msg_waiter.done():
            self.msg_waiter.set_exception(errors.ClientConnectionClosedError())
            self.msg_waiter = None

        if self.transport is not None:
            # With asyncio sslproto on CPython 3.10 or lower, a normal exit
            # (connection closed by peer) cannot set the transport._closed
            # properly, leading to false ResourceWarning. Let's fix that by
            # closing the transport again.
            if not self.transport.is_closing():
                self.transport.close()
            self.transport = None

    def pause_writing(self):
        pass

    def resume_writing(self):
        pass

    def data_received(self, data):
        self.buffer.feed_data(data)

        if (self.msg_waiter is not None and
                self.buffer.take_message() and
                not self.msg_waiter.done()):
            self.msg_waiter.set_result(True)
            self.msg_waiter = None

    def eof_received(self):
        pass



import asyncio
import time

from . import errors
from . import transaction as _transaction


class AsyncIOIteration(_transaction.BaseAsyncIOTransaction):
    def __init__(self, retry, owner, iteration):
        super().__init__(owner, retry._options.transaction_options)
        self.__retry = retry
        self.__iteration = iteration
        self.__started = False

    async def _ensure_transaction(self):
        if not self._managed:
            raise errors.InterfaceError(
                "Only managed retriable transactions are supported. "
                "Use `async with transaction:`"
            )
        if not self.__started:
            self.__started = True
            await self._start(single_connect=self.__iteration != 0)
            if self._pool is not None:
                # Having a pool means we just acquired the connection in
                # _start() - let's mark it as borrowed for transaction anyways
                # just in case the connection is somehow accessed separately.
                self._borrow()

    async def __aenter__(self):
        if self._managed:
            raise errors.InterfaceError(
                'cannot enter context: already in an `async with` block')
        self._managed = True
        if self._pool is None:
            # Borrow the connection for transaction now if it's not on a pool,
            # because that means we already have the connection now, and
            # further use of the connection like this should be prevented:
            #     async for tx in conn.transaction():
            #         async with tx:
            #             await conn.query("...")  # <- wrong use after borrow
            self._borrow()
        return self

    async def __aexit__(self, extype, ex, tb):
        self._managed = False
        if not self.__started:
            self._maybe_return()
            return False

        try:
            if extype is not None:
                await self._rollback()
            else:
                await self._commit()
        except errors.EdgeDBError as err:
            if ex is None:
                # On commit we don't know if commit is succeeded before the
                # database have received it or after it have been done but
                # network is dropped before we were able to receive a response
                # TODO(tailhook) retry on some errors
                raise err
            # If we were going to rollback, look at original error
            # to find out whether we want to retry, regardless of
            # the rollback error.
            # In this case we ignore rollback issue as original error is more
            # important, e.g. in case `CancelledError` it's important
            # to propagate it to cancel the whole task.
            # NOTE: rollback error is always swallowed, should we use
            # on_log_message for it?

        if (
            extype is not None and
            issubclass(extype, errors.EdgeDBError) and
            ex.has_tag(errors.SHOULD_RETRY)
        ):
            return self.__retry._retry(ex)


class BaseRetry:

    def __init__(self, owner):
        self._owner = owner
        self._iteration = 0
        self._done = False
        self._next_backoff = 0
        self._options = owner._options

    def _retry(self, exc):
        self._last_exception = exc
        rule = self._options.retry_options.get_rule_for_exception(exc)
        if self._iteration >= rule.attempts:
            return False
        self._done = False
        self._next_backoff = rule.backoff(self._iteration)
        return True


class AsyncIORetry(BaseRetry):

    def __aiter__(self):
        return self

    async def __anext__(self):
        # Note: when changing this code consider also
        # updating Retry.__next__.
        if self._done:
            raise StopAsyncIteration
        if self._next_backoff:
            await asyncio.sleep(self._next_backoff)
        self._done = True
        iteration = AsyncIOIteration(self, self._owner, self._iteration)
        self._iteration += 1
        return iteration


class Retry(BaseRetry):

    def __iter__(self):
        return self

    def __next__(self):
        # Note: when changing this code consider also
        # updating AsyncIORetry.__anext__.
        if self._done:
            raise StopIteration
        if self._next_backoff:
            time.sleep(self._next_backoff)
        self._done = True
        iteration = Iteration(self, self._owner, self._iteration)
        self._iteration += 1
        return iteration


class Iteration(_transaction.BaseBlockingIOTransaction):
    def __init__(self, retry, owner, iteration):
        super().__init__(owner, retry._options.transaction_options)
        self.__retry = retry
        self.__iteration = iteration
        self.__started = False

    def _ensure_transaction(self):
        if not self._managed:
            raise errors.InterfaceError(
                "Only managed retriable transactions are supported. "
                "Use `with transaction:`"
            )
        if not self.__started:
            self.__started = True
            self._start(single_connect=self.__iteration != 0)

    def __enter__(self):
        if self._managed:
            raise errors.InterfaceError(
                'cannot enter context: already in a `with` block')
        self._managed = True
        self._borrow()
        return self

    def __exit__(self, extype, ex, tb):
        self._managed = False
        if not self.__started:
            self._maybe_return()
            return False

        try:
            if extype is not None:
                self._rollback()
            else:
                self._commit()
        except errors.EdgeDBError as err:
            if ex is None:
                # On commit we don't know if commit is succeeded before the
                # database have received it or after it have been done but
                # network is dropped before we were able to receive a response
                # TODO(tailhook) retry on some errors
                raise err
            # If we were going to rollback, look at original error
            # to find out whether we want to retry, regardless of
            # the rollback error.
            # In this case we ignore rollback issue as original error is more
            # important.

        if (
            extype is not None and
            issubclass(extype, errors.EdgeDBError) and
            ex.has_tag(errors.SHOULD_RETRY)
        ):
            return self.__retry._retry(ex)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


@cython.final
cdef class SetCodec(BaseArrayCodec):

    cdef _new_collection(self, Py_ssize_t size):
        return datatypes.set_new(size)

    cdef _set_collection_item(self, object collection, Py_ssize_t i,
                              object element):
        datatypes.set_set(collection, i, element)

    @staticmethod
    cdef BaseCodec new(bytes tid, BaseCodec sub_codec):
        cdef:
            SetCodec codec

        codec = SetCodec.__new__(SetCodec)

        codec.tid = tid
        codec.name = 'Set'
        codec.sub_codec = sub_codec

        return codec

    cdef decode(self, FRBuffer *buf):
        if type(self.sub_codec) is ArrayCodec:
            # This is a set of arrays encoded as an array
            # of single-element records.
            return self._decode_array_set(buf)
        else:
            # Set of non-arrays.
            return self._decode_array(buf)

    cdef inline _decode_array_set(self, FRBuffer *buf):
        cdef:
            object result
            object elem
            Py_ssize_t elem_count
            Py_ssize_t recsize
            Py_ssize_t i
            int32_t elem_len
            int32_t ndims = hton.unpack_int32(frb_read(buf, 4))
            BaseCodec sub_codec = <BaseCodec>self.sub_codec
            FRBuffer elem_buf

        frb_read(buf, 4)  # ignore flags
        frb_read(buf, 4)  # ignore reserved

        if ndims == 0:
            # Special case for an empty set.
            return self._new_collection(0)
        elif ndims > 1:
            raise RuntimeError('expected a two-dimensional array for a '
                               'set of arrays')

        elem_count = <Py_ssize_t><uint32_t>hton.unpack_int32(frb_read(buf, 4))
        frb_read(buf, 4)  # Ignore the lower bound information

        result = self._new_collection(elem_count)
        for i in range(elem_count):
            frb_read(buf, 4)  # ignore array element size

            recsize = <Py_ssize_t><uint32_t>hton.unpack_int32(frb_read(buf, 4))
            if recsize != 1:
                raise RuntimeError(
                    'expected a record with a single element as an array set '
                    'element envelope')

            frb_read(buf, 4)  # reserved

            elem_len = hton.unpack_int32(frb_read(buf, 4))
            if elem_len == -1:
                raise RuntimeError(
                    'unexpected NULL value in array set element ')

            frb_slice_from(&elem_buf, buf, elem_len)
            elem = sub_codec.decode(&elem_buf)
            if frb_get_len(&elem_buf):
                raise RuntimeError(
                    f'unexpected trailing data in buffer after '
                    f'set element decoding: {frb_get_len(&elem_buf)}')
            self._set_collection_item(result, i, elem)

        return result



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# IMPORTANT: this private API is subject to change.


import enum
import functools
import typing

from edgedb.datatypes import datatypes as dt


class PointerKind(enum.Enum):

    LINK = 1
    PROPERTY = 2


class PointerDescription(typing.NamedTuple):

    name: str
    kind: PointerKind
    implicit: bool


class ObjectDescription(typing.NamedTuple):

    pointers: typing.Tuple[PointerDescription, ...]


@functools.lru_cache()
def _introspect_object_desc(desc) -> ObjectDescription:
    pointers = []
    # Call __dir__ directly as dir() scrambles the order.
    for name in desc.__dir__():
        if desc.is_link(name):
            kind = PointerKind.LINK
        elif desc.is_linkprop(name):
            continue
        else:
            kind = PointerKind.PROPERTY

        pointers.append(
            PointerDescription(
                name=name,
                kind=kind,
                implicit=desc.is_implicit(name)))

    return ObjectDescription(
        pointers=tuple(pointers))


def introspect_object(obj) -> ObjectDescription:
    return _introspect_object_desc(
        dt.get_object_descriptor(obj))



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from libc.stdint cimport int64_t


DEF KiB = 1024;
DEF MiB = 1024 * KiB;
DEF GiB = 1024 * MiB;
DEF TiB = 1024 * GiB;
DEF PiB = 1024 * TiB;


@cython.final
cdef class ConfigMemory:

    def __init__(self, int64_t bytes):
        self._bytes = bytes

    def __eq__(self, other):
        if type(other) is not ConfigMemory:
            return NotImplemented

        return self._bytes == (<ConfigMemory>other)._bytes

    def __hash__(self):
        return hash((ConfigMemory, self._bytes))

    def __repr__(self):
        return f'<edgedb.ConfigMemory "{self}">'

    @cython.cdivision(True)
    def __str__(self):
        cdef:
            int64_t bytes = self._bytes

        if bytes >= PiB and bytes % PiB == 0:
            return f'{bytes // PiB}PiB'
        if bytes >= TiB and bytes % TiB == 0:
            return f'{bytes // TiB}TiB'
        if bytes >= GiB and bytes % GiB == 0:
            return f'{bytes // GiB}GiB'
        if bytes >= MiB and bytes % MiB == 0:
            return f'{bytes // MiB}MiB'
        if bytes >= KiB and bytes % KiB == 0:
            return f'{bytes // KiB}KiB'
        return f'{bytes}B'

    def as_bytes(self):
        return self._bytes


cdef new_config_memory(int64_t bytes):
    cdef ConfigMemory mem = ConfigMemory.__new__(ConfigMemory)

    mem._bytes = bytes

    return mem



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import functools
import random
import socket
import ssl
import time
import typing
import warnings

from . import abstract
from . import base_con
from . import compat
from . import con_utils
from . import errors
from . import enums
from . import options
from . import retry as _retry
from . import transaction as _transaction

from .datatypes import datatypes
from .protocol import asyncio_proto
from .protocol import protocol
from .protocol.protocol import CodecsRegistry as _CodecsRegistry
from .protocol.protocol import QueryCodecsCache as _QueryCodecsCache


class _AsyncIOConnectionImpl:

    def __init__(self, codecs_registry, query_cache):
        self._addr = None
        self._transport = None
        self._protocol = None
        self._codecs_registry = codecs_registry
        self._query_cache = query_cache

    def is_closed(self):
        protocol = self._protocol
        return protocol is None or not protocol.connected

    async def connect(self, loop, addrs, config, params, *,
                      single_attempt=False, connection):
        addr = None
        start = time.monotonic()
        if single_attempt:
            max_time = 0
        else:
            max_time = start + config.wait_until_available
        iteration = 1

        while True:
            for addr in addrs:
                try:
                    await compat.wait_for(
                        self._connect_addr(loop, addr, params, connection),
                        config.connect_timeout,
                    )
                except asyncio.TimeoutError as e:
                    if iteration == 1 or time.monotonic() < max_time:
                        continue
                    else:
                        raise errors.ClientConnectionTimeoutError(
                            f"connecting to {addr} failed in"
                            f" {config.connect_timeout} sec"
                        ) from e
                except errors.ClientConnectionError as e:
                    if (
                        e.has_tag(errors.SHOULD_RECONNECT) and
                        (iteration == 1 or time.monotonic() < max_time)
                    ):
                        continue
                    nice_err = e.__class__(
                        con_utils.render_client_no_connection_error(
                            e,
                            addr,
                            attempts=iteration,
                            duration=time.monotonic() - start,
                        ))
                    raise nice_err from e.__cause__
                else:
                    return

            iteration += 1
            await asyncio.sleep(0.01 + random.random() * 0.2)

    async def _connect_addr(self, loop, addr, params, connection):

        factory = functools.partial(
            asyncio_proto.AsyncIOProtocol, params, loop
        )
        tr = None

        try:
            if isinstance(addr, str):
                # UNIX socket
                tr, pr = await loop.create_unix_connection(factory, addr)
            else:
                try:
                    tr, pr = await loop.create_connection(
                        factory, *addr, ssl=params.ssl_ctx
                    )
                except ssl.CertificateError as e:
                    raise con_utils.wrap_error(e) from e
                except ssl.SSLError as e:
                    if e.reason == 'CERTIFICATE_VERIFY_FAILED':
                        raise con_utils.wrap_error(e) from e
                    tr, pr = await loop.create_connection(
                        functools.partial(factory, tls_compat=True), *addr
                    )
                else:
                    con_utils.check_alpn_protocol(
                        tr.get_extra_info('ssl_object')
                    )
        except socket.gaierror as e:
            # All name resolution errors are considered temporary
            raise errors.ClientConnectionFailedTemporarilyError(str(e)) from e
        except OSError as e:
            raise con_utils.wrap_error(e) from e
        except Exception:
            if tr is not None:
                tr.close()
            raise

        pr.set_connection(connection._inner)

        try:
            await pr.connect()
        except OSError as e:
            if tr is not None:
                tr.close()
            raise con_utils.wrap_error(e) from e
        except Exception:
            if tr is not None:
                tr.close()
            raise

        self._transport = tr
        self._protocol = pr
        self._addr = addr

    async def privileged_execute(self, query):
        await self._protocol.simple_query(query, enums.Capability.ALL)

    async def aclose(self):
        """Send graceful termination message wait for connection to drop."""
        if not self.is_closed():
            try:
                self._protocol.terminate()
                await self._protocol.wait_for_disconnect()
            except (Exception, asyncio.CancelledError):
                self.terminate()
                raise

    def terminate(self):
        if not self.is_closed():
            self._protocol.abort()


class _AsyncIOInnerConnection(base_con._InnerConnection):

    def __init__(self, loop, addrs, config, params, *,
                 codecs_registry=None, query_cache=None):
        super().__init__(
            addrs, config, params,
            codecs_registry=codecs_registry, query_cache=query_cache)
        self._loop = loop

    def _detach(self):
        impl = self._impl
        self._impl = None
        new_conn = self.__class__(
            self._loop, self._addrs, self._config, self._params,
            codecs_registry=self._codecs_registry,
            query_cache=self._query_cache)
        new_conn._impl = impl
        impl._protocol.set_connection(new_conn)
        return new_conn

    def _dispatch_log_message(self, msg):
        for cb in self._log_listeners:
            self._loop.call_soon(cb, self, msg)


class AsyncIOConnection(
    base_con.BaseConnection,
    abstract.AsyncIOExecutor,
    options._OptionsMixin,
):

    def __init__(self, loop, addrs, config, params, *,
                 codecs_registry, query_cache):
        self._inner = _AsyncIOInnerConnection(
            loop, addrs, config, params,
            codecs_registry=codecs_registry,
            query_cache=query_cache)
        super().__init__()

    def _shallow_clone(self):
        if self._inner._borrowed_for:
            raise base_con.borrow_error(self._inner._borrowed_for)
        new_conn = self.__class__.__new__(self.__class__)
        new_conn._inner = self._inner
        return new_conn

    def __repr__(self):
        if self.is_closed():
            return '<{classname} [closed] {id:#x}>'.format(
                classname=self.__class__.__name__, id=id(self))
        else:
            return '<{classname} [connected to {addr}] {id:#x}>'.format(
                classname=self.__class__.__name__,
                addr=self.connected_addr(),
                id=id(self))

    async def ensure_connected(self, *, single_attempt=False):
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect(single_attempt=single_attempt)

    # overriden by connection pool
    async def _reconnect(self, single_attempt=False):
        inner = self._inner
        impl = _AsyncIOConnectionImpl(
            inner._codecs_registry, inner._query_cache)
        await impl.connect(inner._loop, inner._addrs,
                           inner._config, inner._params,
                           single_attempt=single_attempt,
                           connection=self)
        inner._impl = impl

    async def _fetchall(
        self,
        query: str,
        *args,
        __limit__: int=0,
        __typeids__: bool=False,
        __typenames__: bool=False,
        __allow_capabilities__: typing.Optional[int]=None,
        **kwargs,
    ) -> datatypes.Set:
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()
        result, _ = await inner._impl._protocol.execute_anonymous(
            query=query,
            args=args,
            kwargs=kwargs,
            reg=inner._codecs_registry,
            qc=inner._query_cache,
            implicit_limit=__limit__,
            inline_typeids=__typeids__,
            inline_typenames=__typenames__,
            io_format=protocol.IoFormat.BINARY,
            allow_capabilities=__allow_capabilities__,
        )
        return result

    async def _fetchall_with_headers(
        self,
        query: str,
        *args,
        __limit__: int=0,
        __typeids__: bool=False,
        __typenames__: bool=False,
        __allow_capabilities__: typing.Optional[int]=None,
        **kwargs,
    ) -> typing.Tuple[datatypes.Set, typing.Dict[int, bytes]]:
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()
        return await inner._impl._protocol.execute_anonymous(
            query=query,
            args=args,
            kwargs=kwargs,
            reg=inner._codecs_registry,
            qc=inner._query_cache,
            implicit_limit=__limit__,
            inline_typeids=__typeids__,
            inline_typenames=__typenames__,
            io_format=protocol.IoFormat.BINARY,
            allow_capabilities=__allow_capabilities__,
        )

    async def _fetchall_json(
        self,
        query: str,
        *args,
        __limit__: int=0,
        **kwargs,
    ) -> datatypes.Set:
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()
        result, _ = await inner._impl._protocol.execute_anonymous(
            query=query,
            args=args,
            kwargs=kwargs,
            reg=inner._codecs_registry,
            qc=inner._query_cache,
            implicit_limit=__limit__,
            inline_typenames=False,
            io_format=protocol.IoFormat.JSON,
        )
        return result

    async def _execute(
        self,
        query: str,
        args,
        kwargs,
        io_format,
        expect_one=False,
        required_one=False,
    ):
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()

        reconnect = False
        capabilities = None
        i = 0
        while True:
            i += 1
            try:
                if reconnect:
                    await self._reconnect(single_attempt=True)
                result, _ = \
                    await self._inner._impl._protocol.execute_anonymous(
                        query=query,
                        args=args,
                        kwargs=kwargs,
                        reg=inner._codecs_registry,
                        qc=inner._query_cache,
                        io_format=io_format,
                        expect_one=expect_one,
                        required_one=required_one,
                        allow_capabilities=enums.Capability.EXECUTE,
                    )
                return result
            except errors.EdgeDBError as e:
                if not e.has_tag(errors.SHOULD_RETRY):
                    raise e
                if capabilities is None:
                    _, _, _, capabilities = inner._query_cache.get(
                        query=query,
                        io_format=io_format,
                        implicit_limit=0,
                        inline_typenames=False,
                        inline_typeids=False,
                        expect_one=expect_one,
                    )
                # A query is read-only if it has no capabilities i.e.
                # capabilities == 0. Read-only queries are safe to retry.
                if capabilities != 0:
                    raise e
                rule = self._options.retry_options.get_rule_for_exception(e)
                if i >= rule.attempts:
                    raise e
                await asyncio.sleep(rule.backoff(i))
                reconnect = self.is_closed()

    async def query(self, query: str, *args, **kwargs) -> datatypes.Set:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            io_format=protocol.IoFormat.BINARY,
        )

    async def query_single(
        self, query: str, *args, **kwargs
    ) -> typing.Union[typing.Any, None]:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            expect_one=True,
            io_format=protocol.IoFormat.BINARY,
        )

    async def query_required_single(
        self, query: str, *args, **kwargs
    ) -> typing.Any:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            expect_one=True,
            required_one=True,
            io_format=protocol.IoFormat.BINARY,
        )

    async def query_json(self, query: str, *args, **kwargs) -> str:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            io_format=protocol.IoFormat.JSON,
        )

    async def _fetchall_json_elements(
            self, query: str, *args, **kwargs) -> typing.List[str]:
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()
        result, _ = await inner._impl._protocol.execute_anonymous(
            query=query,
            args=args,
            kwargs=kwargs,
            reg=inner._codecs_registry,
            qc=inner._query_cache,
            io_format=protocol.IoFormat.JSON_ELEMENTS,
            allow_capabilities=enums.Capability.EXECUTE,
        )
        return result

    async def query_single_json(self, query: str, *args, **kwargs) -> str:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            io_format=protocol.IoFormat.JSON,
            expect_one=True,
        )

    async def query_required_single_json(
        self, query: str, *args, **kwargs
    ) -> str:
        return await self._execute(
            query=query,
            args=args,
            kwargs=kwargs,
            io_format=protocol.IoFormat.JSON,
            expect_one=True,
            required_one=True
        )

    async def execute(self, query: str) -> None:
        """Execute an EdgeQL command (or commands).

        Example:

        .. code-block:: pycon

            >>> await con.execute('''
            ...     CREATE TYPE MyType { CREATE PROPERTY a -> int64 };
            ...     FOR x IN {100, 200, 300} UNION INSERT MyType { a := x };
            ... ''')
        """
        inner = self._inner
        if inner._borrowed_for:
            raise base_con.borrow_error(inner._borrowed_for)
        if not inner._impl or inner._impl.is_closed():
            await self._reconnect()
        await inner._impl._protocol.simple_query(
            query, enums.Capability.EXECUTE)

    def transaction(self) -> _retry.AsyncIORetry:
        return _retry.AsyncIORetry(self)

    def retrying_transaction(self) -> _retry.AsyncIORetry:
        warnings.warn(
            'The "retrying_transaction()" method has been renamed to '
            '"transaction()"',
            DeprecationWarning, 2)
        return _retry.AsyncIORetry(self)

    def raw_transaction(self) -> _transaction.AsyncIOTransaction:
        warnings.warn(
            'The "raw_transaction()" method is deprecated and is scheduled '
            'to be removed. Use the "transaction()" method with '
            'retry attempts=1 instead',
            DeprecationWarning, 2)
        return _transaction.AsyncIOTransaction(
            self,
            self._options.transaction_options,
        )

    async def aclose(self) -> None:
        try:
            await self._inner._impl.aclose()
        finally:
            self._cleanup()

    def terminate(self) -> None:
        try:
            self._inner._impl.terminate()
        finally:
            self._cleanup()

    def _set_proxy(self, proxy):
        if self._proxy is not None and proxy is not None:
            # Should not happen unless there is a bug in `Pool`.
            raise errors.InterfaceError(
                'internal client error: connection is already proxied')

        self._proxy = proxy

    def is_closed(self) -> bool:
        return self._inner._impl.is_closed()

    async def fetchall(self, query: str, *args, **kwargs) -> datatypes.Set:
        warnings.warn(
            'The "fetchall()" method is deprecated and is scheduled to be '
            'removed. Use the "query()" method instead.',
            DeprecationWarning, 2)
        return await self.query(query, *args, **kwargs)

    async def query_one(self, query: str, *args, **kwargs) -> typing.Any:
        warnings.warn(
            'The "query_one()" method is deprecated and is scheduled to be '
            'removed. Use the "query_single()" method instead.',
            DeprecationWarning, 2)
        return await self.query_single(query, *args, **kwargs)

    async def fetchone(self, query: str, *args, **kwargs) -> typing.Any:
        warnings.warn(
            'The "fetchone()" method is deprecated and is scheduled to be '
            'removed. Use the "query_single()" method instead.',
            DeprecationWarning, 2)
        return await self.query_single(query, *args, **kwargs)

    async def fetchall_json(self, query: str, *args, **kwargs) -> str:
        warnings.warn(
            'The "fetchall_json()" method is deprecated and is scheduled to '
            'be removed. Use the "query_json()" method instead.',
            DeprecationWarning, 2)
        return await self.query_json(query, *args, **kwargs)

    async def query_one_json(self, query: str, *args, **kwargs) -> str:
        warnings.warn(
            'The "query_one_json()" method is deprecated and is scheduled to '
            'be removed. Use the "query_single_json()" method instead.',
            DeprecationWarning, 2)
        return await self.query_single_json(query, *args, **kwargs)

    async def fetchone_json(self, query: str, *args, **kwargs) -> str:
        warnings.warn(
            'The "fetchone_json()" method is deprecated and is scheduled to '
            'be removed. Use the "query_single_json()" method instead.',
            DeprecationWarning, 2)
        return await self.query_single_json(query, *args, **kwargs)


async def async_connect_raw(dsn: str = None, *,
                            credentials_file: str = None,
                            host: str = None, port: int = None,
                            user: str = None, password: str = None,
                            database: str = None,
                            tls_ca_file: str = None,
                            tls_security: str = None,
                            connection_class=None,
                            wait_until_available: int = 30,
                            timeout: int = 10) -> AsyncIOConnection:

    loop = asyncio.get_event_loop()

    if connection_class is None:
        connection_class = AsyncIOConnection

    connect_config, client_config = con_utils.parse_connect_arguments(
        dsn=dsn, credentials_file=credentials_file, host=host, port=port,
        user=user, password=password, database=database, timeout=timeout,
        tls_ca_file=tls_ca_file, tls_security=tls_security,
        wait_until_available=wait_until_available,

        # ToDos
        command_timeout=None,
        server_settings=None)

    connection = connection_class(
        loop, [connect_config.address], client_config, connect_config,
        codecs_registry=_CodecsRegistry(),
        query_cache=_QueryCodecsCache(),
    )
    await connection.ensure_connected()
    return connection


async def _connect_addr(loop, addrs, config, params,
                        query_cache, codecs_registry, connection_class):

    if connection_class is None:
        connection_class = AsyncIOConnection

    connection = connection_class(
        loop, addrs, config, params,
        codecs_registry=codecs_registry,
        query_cache=query_cache,
    )
    await connection.ensure_connected()
    return connection



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import functools

from . import errors


def guarded(meth):
    """A decorator to add a sanity check to ConnectionResource methods."""

    @functools.wraps(meth)
    def _check(self, *args, **kwargs):
        self._check_conn_validity(meth.__name__)
        return meth(self, *args, **kwargs)

    return _check


class ConnectionResource:

    def __init__(self, connection):
        self._connection = connection
        self._con_release_ctr = connection._pool_release_ctr

    def _check_conn_validity(self, meth_name):
        con_release_ctr = self._connection._pool_release_ctr
        if con_release_ctr != self._con_release_ctr:
            raise errors.InterfaceError(
                'cannot call {}.{}(): '
                'the underlying connection has been released back '
                'to the pool'.format(self.__class__.__name__, meth_name))

        if self._connection.is_closed():
            raise errors.InterfaceError(
                'cannot call {}.{}(): '
                'the underlying connection is closed'.format(
                    self.__class__.__name__, meth_name))



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio

import logging
import unittest.mock

import edgedb
from edgedb import compat
from edgedb import errors
from edgedb import RetryOptions
from edgedb import _testbase as tb

log = logging.getLogger(__name__)


class Barrier:
    def __init__(self, number):
        self._counter = number
        self._cond = asyncio.Condition()

    async def ready(self):
        if self._counter == 0:
            return
        async with self._cond:
            self._counter -= 1
            assert self._counter >= 0, self._counter
            if self._counter == 0:
                self._cond.notify_all()
            else:
                await self._cond.wait_for(lambda: self._counter == 0)


class TestAsyncRetry(tb.AsyncQueryTestCase):

    ISOLATED_METHODS = False

    SETUP = '''
        CREATE TYPE test::Counter EXTENDING std::Object {
            CREATE PROPERTY name -> std::str {
                CREATE CONSTRAINT std::exclusive;
            };
            CREATE PROPERTY value -> std::int32 {
                SET default := 0;
            };
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::Counter;
    '''

    async def test_async_retry_01(self):
        async for tx in self.con.transaction():
            async with tx:
                await tx.execute('''
                    INSERT test::Counter {
                        name := 'counter1'
                    };
                ''')

    async def test_async_retry_02(self):
        with self.assertRaises(ZeroDivisionError):
            async for tx in self.con.transaction():
                async with tx:
                    await tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_02'
                        };
                    ''')
                    1 / 0
        with self.assertRaises(edgedb.NoDataError):
            await self.con.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_02'
            ''')

    async def test_async_retry_begin(self):
        patcher = unittest.mock.patch("edgedb.retry.AsyncIOIteration._start")
        _start = patcher.start()

        def cleanup():
            try:
                patcher.stop()
            except RuntimeError:
                pass

        self.addCleanup(cleanup)

        _start.side_effect = errors.BackendUnavailableError()

        with self.assertRaises(errors.BackendUnavailableError):
            async for tx in self.con.transaction():
                async with tx:
                    await tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_begin'
                        };
                    ''')
        with self.assertRaises(edgedb.NoDataError):
            await self.con.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_begin'
            ''')

        async def recover_after_first_error(*_, **__):
            patcher.stop()
            raise errors.BackendUnavailableError()

        _start.side_effect = recover_after_first_error
        call_count = _start.call_count

        async for tx in self.con.transaction():
            async with tx:
                await tx.execute('''
                    INSERT test::Counter {
                        name := 'counter_retry_begin'
                    };
                ''')
        self.assertEqual(_start.call_count, call_count + 1)
        await self.con.query_single('''
            SELECT test::Counter
            FILTER .name = 'counter_retry_begin'
        ''')

    @unittest.skip('https://github.com/edgedb/edgedb/issues/2869')
    async def test_async_retry_conflict(self):
        await self.execute_conflict('counter2')

    @unittest.skip('https://github.com/edgedb/edgedb/issues/2869')
    async def test_async_conflict_no_retry(self):
        with self.assertRaises(edgedb.TransactionSerializationError):
            await self.execute_conflict(
                'counter3',
                RetryOptions(attempts=1, backoff=edgedb.default_backoff)
            )

    async def execute_conflict(self, name='counter2', options=None):
        con2 = await self.connect(database=self.get_database_name())
        self.addCleanup(con2.aclose)

        barrier = Barrier(2)
        lock = asyncio.Lock()
        iterations = 0

        async def transaction1(con):
            async for tx in con.transaction():
                nonlocal iterations
                iterations += 1
                async with tx:
                    # This magic query makes the test more reliable for some
                    # reason. I guess this is because starting a transaction
                    # in EdgeDB (and/or Postgres) is accomplished somewhat
                    # lazily, i.e. only start transaction on the first query
                    # rather than on the `START TRANSACTION`.
                    await tx.query("SELECT 1")

                    # Start both transactions at the same initial data.
                    # One should succeed other should fail and retry.
                    # On next attempt, the latter should succeed
                    await barrier.ready()

                    await lock.acquire()
                    res = await tx.query_single('''
                        SELECT (
                            INSERT test::Counter {
                                name := <str>$name,
                                value := 1,
                            } UNLESS CONFLICT ON .name
                            ELSE (
                                UPDATE test::Counter
                                SET { value := .value + 1 }
                            )
                        ).value
                    ''', name=name)
                lock.release()
            return res

        con = self.con
        if options:
            con = con.with_retry_options(options)
            con2 = con2.with_retry_options(options)

        results = await compat.wait_for(asyncio.gather(
            transaction1(con),
            transaction1(con2),
            return_exceptions=True,
        ), 10)
        for e in results:
            if isinstance(e, BaseException):
                raise e

        self.assertEqual(set(results), {1, 2})
        self.assertEqual(iterations, 3)

    async def test_async_transaction_interface_errors(self):
        with self.assertRaisesRegex(
            AttributeError,
            "'AsyncIOIteration' object has no attribute 'start'",
        ):
            async for tx in self.con.transaction():
                async with tx:
                    await tx.start()

        with self.assertRaisesRegex(
            AttributeError,
            "'AsyncIOIteration' object has no attribute 'rollback'",
        ):
            async for tx in self.con.transaction():
                async with tx:
                    await tx.rollback()

        with self.assertRaisesRegex(
            AttributeError,
            "'AsyncIOIteration' object has no attribute 'start'",
        ):
            async for tx in self.con.transaction():
                await tx.start()

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*Use `async with transaction:`'):
            async for tx in self.con.transaction():
                await tx.execute("SELECT 123")

        with self.assertRaisesRegex(
            edgedb.InterfaceError,
            r"already in an `async with` block",
        ):
            async for tx in self.con.transaction():
                async with tx:
                    async with tx:
                        pass

        with self.assertRaisesRegex(edgedb.InterfaceError, r".*is borrowed.*"):
            async for tx in self.con.transaction():
                async with tx:
                    await self.con.execute("SELECT 123")



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import threading
import unittest.mock
from concurrent import futures

import edgedb
from edgedb import _testbase as tb
from edgedb import errors
from edgedb import RetryOptions


class Barrier:
    def __init__(self, number):
        self._counter = number
        self._cond = threading.Condition()

    def ready(self):
        if self._counter == 0:
            return
        with self._cond:
            self._counter -= 1
            assert self._counter >= 0, self._counter
            if self._counter == 0:
                self._cond.notify_all()
            else:
                self._cond.wait_for(lambda: self._counter == 0)


class TestSyncRetry(tb.SyncQueryTestCase):

    ISOLATED_METHODS = False

    SETUP = '''
        CREATE TYPE test::Counter EXTENDING std::Object {
            CREATE PROPERTY name -> std::str {
                CREATE CONSTRAINT std::exclusive;
            };
            CREATE PROPERTY value -> std::int32 {
                SET default := 0;
            };
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::Counter;
    '''

    def test_sync_retry_01(self):
        for tx in self.con.transaction():
            with tx:
                tx.execute('''
                    INSERT test::Counter {
                        name := 'counter1'
                    };
                ''')

    def test_sync_retry_02(self):
        with self.assertRaises(ZeroDivisionError):
            for tx in self.con.transaction():
                with tx:
                    tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_02'
                        };
                    ''')
                    1 / 0
        with self.assertRaises(edgedb.NoDataError):
            self.con.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_02'
            ''')
        self.assertEqual(
            self.con.query_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_02'
            '''),
            None
        )

    def test_sync_retry_begin(self):
        patcher = unittest.mock.patch("edgedb.retry.Iteration._start")
        _start = patcher.start()

        def cleanup():
            try:
                patcher.stop()
            except RuntimeError:
                pass

        self.addCleanup(cleanup)

        _start.side_effect = errors.BackendUnavailableError()

        with self.assertRaises(errors.BackendUnavailableError):
            for tx in self.con.transaction():
                with tx:
                    tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_begin'
                        };
                    ''')
        with self.assertRaises(edgedb.NoDataError):
            self.con.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_begin'
            ''')
        self.assertEqual(
            self.con.query_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_begin'
            '''),
            None
        )

        def recover_after_first_error(*_, **__):
            patcher.stop()
            raise errors.BackendUnavailableError()

        _start.side_effect = recover_after_first_error
        call_count = _start.call_count

        for tx in self.con.transaction():
            with tx:
                tx.execute('''
                    INSERT test::Counter {
                        name := 'counter_retry_begin'
                    };
                ''')
        self.assertEqual(_start.call_count, call_count + 1)
        self.con.query_single('''
            SELECT test::Counter
            FILTER .name = 'counter_retry_begin'
        ''')

    @unittest.skip('https://github.com/edgedb/edgedb/issues/2869')
    def test_sync_retry_conflict(self):
        self.execute_conflict('counter2')

    @unittest.skip('https://github.com/edgedb/edgedb/issues/2869')
    def test_sync_conflict_no_retry(self):
        with self.assertRaises(edgedb.TransactionSerializationError):
            self.execute_conflict(
                'counter3',
                RetryOptions(attempts=1, backoff=edgedb.default_backoff)
            )

    def execute_conflict(self, name='counter2', options=None):
        con_args = self.get_connect_args().copy()
        con_args.update(database=self.get_database_name())
        con2 = edgedb.connect(**con_args)
        self.addCleanup(con2.close)

        barrier = Barrier(2)
        lock = threading.Lock()

        iterations = 0

        def transaction1(con):
            for tx in con.transaction():
                nonlocal iterations
                iterations += 1
                with tx:
                    # This magic query makes the test more reliable for some
                    # reason. I guess this is because starting a transaction
                    # in EdgeDB (and/or Postgres) is accomplished somewhat
                    # lazily, i.e. only start transaction on the first query
                    # rather than on the `START TRANSACTION`.
                    tx.query("SELECT 1")

                    # Start both transactions at the same initial data.
                    # One should succeed other should fail and retry.
                    # On next attempt, the latter should succeed
                    barrier.ready()

                    lock.acquire()
                    res = tx.query_single('''
                        SELECT (
                            INSERT test::Counter {
                                name := <str>$name,
                                value := 1,
                            } UNLESS CONFLICT ON .name
                            ELSE (
                                UPDATE test::Counter
                                SET { value := .value + 1 }
                            )
                        ).value
                    ''', name=name)
                lock.release()
            return res

        con = self.con
        if options:
            con = con.with_retry_options(options)
            con2 = con2.with_retry_options(options)

        with futures.ThreadPoolExecutor(2) as pool:
            f1 = pool.submit(transaction1, con)
            f2 = pool.submit(transaction1, con2)
            results = {f1.result(), f2.result()}

        self.assertEqual(results, {1, 2})
        self.assertEqual(iterations, 3)

    def test_sync_transaction_interface_errors(self):
        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'start'",
        ):
            for tx in self.con.transaction():
                with tx:
                    tx.start()

        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'rollback'",
        ):
            for tx in self.con.transaction():
                with tx:
                    tx.rollback()

        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'start'",
        ):
            for tx in self.con.transaction():
                tx.start()

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*Use `with transaction:`'):
            for tx in self.con.transaction():
                tx.execute("SELECT 123")

        with self.assertRaisesRegex(
            edgedb.InterfaceError,
            r"already in a `with` block",
        ):
            for tx in self.con.transaction():
                with tx:
                    with tx:
                        pass

        with self.assertRaisesRegex(edgedb.InterfaceError, r".*is borrowed.*"):
            for tx in self.con.transaction():
                with tx:
                    self.con.execute("SELECT 123")



import enum


class Capability(enum.IntFlag):

    MODIFICATIONS     = 1 << 0    # noqa
    SESSION_CONFIG    = 1 << 1    # noqa
    TRANSACTION       = 1 << 2    # noqa
    DDL               = 1 << 3    # noqa
    PERSISTENT_CONFIG = 1 << 4    # noqa


Capability.ALL = 0xFFFF_FFFF_FFFF_FFFF
Capability.EXECUTE = Capability.ALL & ~Capability.TRANSACTION



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


cimport cython
cimport cpython

from libc.stdint cimport int16_t, int32_t, uint16_t, \
                         uint32_t, int64_t, uint64_t

from edgedb.pgproto.pgproto cimport (
    WriteBuffer,
    ReadBuffer,
    FRBuffer,
)

from edgedb.pgproto cimport pgproto
from edgedb.pgproto.debug cimport PG_DEBUG


include "./lru.pxd"
include "./codecs/codecs.pxd"


ctypedef object (*decode_row_method)(BaseCodec, FRBuffer *buf)


cpdef enum IoFormat:
    BINARY = b'b'
    JSON = b'j'
    JSON_ELEMENTS = b'J'


cdef enum TransactionStatus:
    TRANS_IDLE = 0                 # connection idle
    TRANS_ACTIVE = 1               # command in progress
    TRANS_INTRANS = 2              # idle, within transaction block
    TRANS_INERROR = 3              # idle, within failed transaction
    TRANS_UNKNOWN = 4              # cannot determine status


cdef enum EdgeParseFlags:
    PARSE_HAS_RESULT = 1 << 0
    PARSE_SINGLETON_RESULT = 1 << 1


cdef enum AuthenticationStatuses:
    AUTH_OK = 0
    AUTH_SASL = 10
    AUTH_SASL_CONTINUE = 11
    AUTH_SASL_FINAL = 12


cdef class QueryCodecsCache:

    cdef:
        LRUMapping queries

    cdef set(self, str query, IoFormat io_format,
             int implicit_limit, bint inline_typenames, bint inline_typeids,
             bint expect_one, bint has_na_cardinality,
             BaseCodec in_type, BaseCodec out_type, int capabilities)


cdef class SansIOProtocol:

    cdef:
        ReadBuffer buffer

        readonly bint connected
        readonly bint cancelled

        object con
        readonly object con_params
        readonly bint tls_compat

        object backend_secret

        TransactionStatus xact_status

        CodecsRegistry internal_reg
        dict server_settings

        readonly bytes last_status
        readonly bytes last_details
        readonly tuple protocol_version

    cdef encode_args(self, BaseCodec in_dc, WriteBuffer buf, args, kwargs)

    cdef parse_data_messages(self, BaseCodec out_dc, result)
    cdef parse_sync_message(self)
    cdef parse_command_complete_message(self)
    cdef parse_describe_type_message(self, CodecsRegistry reg)
    cdef _amend_parse_error(
        self, exc, IoFormat io_format, bint expect_one, bint required_one)

    cdef inline ignore_headers(self)
    cdef dict parse_headers(self)
    cdef write_headers(self, WriteBuffer buf, dict headers)

    cdef parse_error_message(self)

    cdef write(self, WriteBuffer buf)
    cpdef abort(self)

    cdef reset_status(self)

    cdef parse_system_config(self, BaseCodec codec, bytes data)
    cdef parse_server_settings(self, str name, bytes val)

    cdef fallthrough(self)

    cdef write_execute_headers(
        self,
        WriteBuffer buf,
        int implicit_limit,
        bint inline_typenames,
        bint inline_typeids,
        uint64_t allow_capabilities,
    )

    cdef ensure_connected(self)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from libc.stdint cimport int64_t, int32_t


@cython.final
cdef class RelativeDuration:

    cdef:
        readonly int64_t microseconds
        readonly int32_t days
        readonly int32_t months


cdef new_duration(int64_t microseconds, int32_t days, int32_t months)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import itertools

import edgedb

from edgedb import _testbase as tb
from edgedb import TransactionOptions


class TestSyncTx(tb.SyncQueryTestCase):

    ISOLATED_METHODS = False

    SETUP = '''
        CREATE TYPE test::TransactionTest EXTENDING std::Object {
            CREATE PROPERTY name -> std::str;
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::TransactionTest;
    '''

    def test_sync_transaction_regular_01(self):
        self.assertIsNone(self.con._inner._borrowed_for)
        tr = self.con.transaction()
        self.assertIsNone(self.con._inner._borrowed_for)

        with self.assertRaises(ZeroDivisionError):
            for with_tr in tr:
                with with_tr:
                    with_tr.execute('''
                        INSERT test::TransactionTest {
                            name := 'Test Transaction'
                        };
                    ''')

                    1 / 0

        self.assertIsNone(self.con._inner._borrowed_for)

        result = self.con.query('''
            SELECT
                test::TransactionTest
            FILTER
                test::TransactionTest.name = 'Test Transaction';
        ''')

        self.assertEqual(result, [])

    async def test_sync_transaction_kinds(self):
        isolations = [
            None,
            edgedb.IsolationLevel.Serializable,
        ]
        booleans = [None, True, False]
        all = itertools.product(isolations, booleans, booleans)
        for isolation, readonly, deferrable in all:
            opt = dict(
                isolation=isolation,
                readonly=readonly,
                deferrable=deferrable,
            )
            # skip None
            opt = {k: v for k, v in opt.items() if v is not None}
            con = self.con.with_transaction_options(TransactionOptions(**opt))
            try:
                for tx in con.transaction():
                    with tx:
                        tx.execute(
                            'INSERT test::TransactionTest {name := "test"}')
            except edgedb.TransactionError:
                self.assertTrue(readonly)
            else:
                self.assertFalse(readonly)

            for tx in con.transaction():
                with tx:
                    pass

    def test_sync_transaction_interface_errors(self):
        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot start; .* already started'):
            with tr:
                tr.start()

        self.assertTrue(repr(tr).startswith(
            '<edgedb.Transaction state:rolledback'))

        self.assertIsNone(self.con._inner._top_xact)

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot start; .* already rolled back'):
            with tr:
                pass

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot manually commit.*with'):
            with tr:
                tr.commit()

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot manually rollback.*with'):
            with tr:
                tr.rollback()

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot enter context:.*with'):
            with tr:
                with tr:
                    pass

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.query("SELECT 1")

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.query_single("SELECT 1")

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.query_json("SELECT 1")

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.query_single_json("SELECT 1")

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.execute("SELECT 1")

        tr = self.con.raw_transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'.*is borrowed.*'):
            with tr:
                self.con.ensure_connected()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unittest


import edgedb
from edgedb.datatypes import datatypes as private
from edgedb import introspect


class TestRecordDesc(unittest.TestCase):

    def test_recorddesc_1(self):
        with self.assertRaisesRegex(TypeError, 'one to three positional'):
            private._RecordDescriptor()

        with self.assertRaisesRegex(TypeError, 'one to three positional'):
            private._RecordDescriptor(t=1)

        with self.assertRaisesRegex(TypeError, 'requires a tuple'):
            private._RecordDescriptor(1)

        with self.assertRaisesRegex(TypeError, 'requires a tuple'):
            private._RecordDescriptor(('a',), 1)

        with self.assertRaisesRegex(TypeError,
                                    'the same length as the names tuple'):
            private._RecordDescriptor(('a',), ())

        private._RecordDescriptor(('a', 'b'))

        with self.assertRaisesRegex(ValueError, f'more than {0x4000-1}'):
            private._RecordDescriptor(('a',) * 20000)

    def test_recorddesc_2(self):
        rd = private._RecordDescriptor(
            ('a', 'b', 'c'),
            (private._EDGE_POINTER_IS_LINKPROP,
             0,
             private._EDGE_POINTER_IS_LINK))

        self.assertEqual(rd.get_pos('a'), 0)
        self.assertEqual(rd.get_pos('b'), 1)
        self.assertEqual(rd.get_pos('c'), 2)

        self.assertTrue(rd.is_linkprop('a'))
        self.assertFalse(rd.is_linkprop('b'))
        self.assertFalse(rd.is_linkprop('c'))

        self.assertFalse(rd.is_link('a'))
        self.assertFalse(rd.is_link('b'))
        self.assertTrue(rd.is_link('c'))

        with self.assertRaises(LookupError):
            rd.get_pos('z')

        with self.assertRaises(LookupError):
            rd.is_linkprop('z')

    def test_recorddesc_3(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property',
            d='link',
        )

        o = f(1, 2, 3, 4)

        desc = private.get_object_descriptor(o)
        self.assertEqual(set(dir(desc)), set(('id', 'lb', 'c', 'd')))

        self.assertTrue(desc.is_linkprop('lb'))
        self.assertFalse(desc.is_linkprop('id'))
        self.assertFalse(desc.is_linkprop('c'))
        self.assertFalse(desc.is_linkprop('d'))

        self.assertFalse(desc.is_link('lb'))
        self.assertFalse(desc.is_link('id'))
        self.assertFalse(desc.is_link('c'))
        self.assertTrue(desc.is_link('d'))

        self.assertFalse(desc.is_implicit('lb'))
        self.assertTrue(desc.is_implicit('id'))
        self.assertFalse(desc.is_implicit('c'))
        self.assertFalse(desc.is_implicit('d'))

        self.assertEqual(desc.get_pos('lb'), 1)
        self.assertEqual(desc.get_pos('id'), 0)
        self.assertEqual(desc.get_pos('c'), 2)
        self.assertEqual(desc.get_pos('d'), 3)

    def test_recorddesc_4(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property',
            d='link',
        )

        o = f(1, 2, 3, 4)
        intro = introspect.introspect_object(o)

        self.assertEqual(
            intro.pointers,
            (
                ('id', introspect.PointerKind.PROPERTY, True),
                ('c', introspect.PointerKind.PROPERTY, False),
                ('d', introspect.PointerKind.LINK, False),
            )
        )

        # clear cache so that tests in refcount mode don't freak out.
        introspect._introspect_object_desc.cache_clear()


class TestTuple(unittest.TestCase):

    def test_tuple_empty_1(self):
        t = edgedb.Tuple()
        self.assertEqual(len(t), 0)
        self.assertEqual(hash(t), hash(()))
        self.assertEqual(repr(t), '()')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[0]

    def test_tuple_2(self):
        t = edgedb.Tuple((1, 'a'))
        self.assertEqual(len(t), 2)
        self.assertEqual(hash(t), hash((1, 'a')))

        self.assertEqual(repr(t), "(1, 'a')")

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

    def test_tuple_3(self):
        t = edgedb.Tuple((1, []))
        t[1].append(t)
        self.assertEqual(t[1], [t])

        self.assertEqual(repr(t), '(1, [(...)])')
        self.assertEqual(str(t), '(1, [(...)])')

    def test_tuple_4(self):
        with self.assertRaisesRegex(ValueError, f'more than {0x4000 - 1}'):
            edgedb.Tuple([1] * 20000)

    def test_tuple_freelist_1(self):
        lst = []
        for _ in range(5000):
            lst.append(edgedb.Tuple((1,)))
        for t in lst:
            self.assertEqual(t[0], 1)

    def test_tuple_5(self):
        self.assertEqual(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 3, 2]))

        self.assertLess(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 3, 2]))

        self.assertEqual(
            edgedb.Tuple([]),
            edgedb.Tuple([]))

        self.assertEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([1]))

        self.assertGreaterEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([1]))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([]))

        self.assertGreater(
            edgedb.Tuple([1]),
            edgedb.Tuple([]))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([2]))

        self.assertLess(
            edgedb.Tuple([1]),
            edgedb.Tuple([2]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2]),
            edgedb.Tuple([2, 2]))

        self.assertNotEqual(
            edgedb.Tuple([1, 1]),
            edgedb.Tuple([2, 2, 1]))

    def test_tuple_6(self):
        self.assertEqual(
            edgedb.Tuple([1, 2, 3]),
            (1, 2, 3))

        self.assertEqual(
            (1, 2, 3),
            edgedb.Tuple([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            (1, 3, 2))

        self.assertLess(
            edgedb.Tuple([1, 2, 3]),
            (1, 3, 2))

        self.assertEqual(
            edgedb.Tuple([]),
            ())

        self.assertEqual(
            edgedb.Tuple([1]),
            (1,))

        self.assertGreaterEqual(
            edgedb.Tuple([1]),
            (1,))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            ())

        self.assertGreater(
            edgedb.Tuple([1]),
            ())

        self.assertNotEqual(
            edgedb.Tuple([1]),
            (2,))

        self.assertLess(
            edgedb.Tuple([1]),
            (2,))

        self.assertNotEqual(
            edgedb.Tuple([1, 2]),
            (2, 2))

        self.assertNotEqual(
            edgedb.Tuple([1, 1]),
            (2, 2, 1))

    def test_tuple_7(self):
        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            123)


class TestNamedTuple(unittest.TestCase):

    def test_namedtuple_empty_1(self):
        with self.assertRaisesRegex(ValueError, 'at least one field'):
            edgedb.NamedTuple()

    def test_namedtuple_2(self):
        t = edgedb.NamedTuple(a=1)
        self.assertEqual(repr(t), "(a := 1)")

        t = edgedb.NamedTuple(a=1, b='a')

        self.assertEqual(set(dir(t)), {'a', 'b'})

        self.assertEqual(repr(t), "(a := 1, b := 'a')")

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

        self.assertEqual(len(t), 2)
        self.assertEqual(hash(t), hash((1, 'a')))

        self.assertEqual(t.a, 1)
        self.assertEqual(t.b, 'a')

        with self.assertRaises(AttributeError):
            t.z

    def test_namedtuple_3(self):
        t = edgedb.NamedTuple(a=1, b=[])
        t.b.append(t)
        self.assertEqual(t.b, [t])

        self.assertEqual(repr(t), '(a := 1, b := [(...)])')
        self.assertEqual(str(t), '(a := 1, b := [(...)])')

    def test_namedtuple_4(self):
        t1 = edgedb.NamedTuple(a=1, b='aaaa')
        t2 = edgedb.Tuple((1, 'aaaa'))
        t3 = (1, 'aaaa')

        self.assertEqual(hash(t1), hash(t2))
        self.assertEqual(hash(t1), hash(t3))

    def test_namedtuple_5(self):
        self.assertEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(x=1, y=2, z=3))

        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(a=1, c=3, b=2))

        self.assertLess(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(a=1, b=3, c=2))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            edgedb.NamedTuple(b=1))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            edgedb.NamedTuple(a=1))

    def test_namedtuple_6(self):
        self.assertEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 2, 3))

        self.assertEqual(
            (1, 2, 3),
            edgedb.NamedTuple(a=1, b=2, c=3))

        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 3, 2))

        self.assertLess(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 3, 2))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            (1,))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            (1,))

    def test_namedtuple_7(self):
        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            1)


class TestObject(unittest.TestCase):

    def test_object_1(self):
        f = private.create_object_factory(
            id='property',
            lb='link-property',
            c='property'
        )

        o = f(1, 2, 3)

        self.assertEqual(repr(o), 'Object{id := 1, @lb := 2, c := 3}')

        self.assertEqual(o.id, 1)
        self.assertEqual(o.c, 3)

        with self.assertRaises(AttributeError):
            o.lb

        with self.assertRaises(AttributeError):
            o.z

        with self.assertRaises(TypeError):
            len(o)

        with self.assertRaises(KeyError):
            o[0]

        with self.assertRaises(TypeError):
            o['id']

        self.assertEqual(set(dir(o)), {'id', 'c'})

    def test_object_2(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o = f(1, 2, 3)

        self.assertEqual(repr(o), 'Object{@lb := 2, c := 3}')

        self.assertEqual(hash(o), hash(f(1, 2, 3)))
        self.assertNotEqual(hash(o), hash(f(1, 2, 'aaaa')))
        self.assertNotEqual(hash(o), hash((1, 2, 3)))

        self.assertEqual(set(dir(o)), {'id', 'c'})

    def test_object_3(self):
        f = private.create_object_factory(id='property', c='link')
        o = f(1, [])

        o.c.append(o)
        self.assertEqual(repr(o), 'Object{id := 1, c := [Object{...}]}')

        with self.assertRaisesRegex(TypeError, 'unhashable'):
            hash(o)

    def test_object_4(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o1 = f(1, 'aa', 'ba')
        o2 = f(1, 'ab', 'bb')
        o3 = f(3, 'ac', 'bc')

        self.assertEqual(o1, o2)
        self.assertNotEqual(o1, o3)
        self.assertLess(o1, o3)
        self.assertGreater(o3, o2)

    def test_object_5(self):
        f = private.create_object_factory(
            a='property',
            lb='link-property',
            c='property'
        )
        with self.assertRaisesRegex(ValueError, "without 'id' field"):
            f(1, 2, 3)

    def test_object_6(self):
        User = private.create_object_factory(
            id='property',
            name='property',
        )

        u = User(1, 'user1')

        with self.assertRaisesRegex(TypeError,
                                    "property 'name' should be "
                                    "accessed via dot notation"):
            u['name']

    def test_object_links_1(self):
        O2 = private.create_object_factory(
            id='property',
            lb='link-property',
            c='property'
        )

        O1 = private.create_object_factory(
            id='property',
            o2s='link'
        )

        o2_1 = O2(1, 'linkprop o2 1', 3)
        o2_2 = O2(4, 'linkprop o2 2', 6)
        o1 = O1(2, edgedb.Set((o2_1, o2_2)))

        linkset = o1['o2s']
        self.assertEqual(len(linkset), 2)
        self.assertEqual(linkset, o1['o2s'])
        self.assertEqual(hash(linkset), hash(o1['o2s']))
        self.assertEqual(
            repr(linkset),
            "LinkSet(name='o2s', source_id=2, target_ids={1, 4})")

        link1 = linkset[0]
        self.assertIs(link1.source, o1)
        self.assertIs(link1.target, o2_1)
        self.assertEqual(
            repr(link1),
            "Link(name='o2s', source_id=2, target_id=1)")
        self.assertEqual(set(dir(link1)), {'target', 'source', 'lb'})

        link2 = linkset[1]
        self.assertIs(link2.source, o1)
        self.assertIs(link2.target, o2_2)

        self.assertNotEqual(link1, link2)

        self.assertEqual(list(linkset), [link1, link2])
        self.assertEqual([link for link in linkset], [link1, link2])

        self.assertNotEqual(link1, link2)

        self.assertEqual(link1.lb, 'linkprop o2 1')
        self.assertEqual(link2.lb, 'linkprop o2 2')

        with self.assertRaises(AttributeError):
            link2.aaaa

    def test_object_links_2(self):
        User = private.create_object_factory(
            id='property',
            friends='link',
            enemies='link',
        )

        u1 = User(1, edgedb.Set([]), edgedb.Set([]))
        u2 = User(2, edgedb.Set([]), edgedb.Set([]))
        u3 = User(3, edgedb.Set([]), edgedb.Set([]))
        u4 = User(4, edgedb.Set([u1, u2]), edgedb.Set([u1, u2]))
        u5 = User(5, edgedb.Set([u1, u3]), edgedb.Set([u1, u2]))

        self.assertNotEqual(u4['friends'], u4['enemies'])
        self.assertNotEqual(u4['enemies'], u5['enemies'])

        self.assertEqual(set(dir(u1)), {'id', 'friends', 'enemies'})

    def test_object_links_3(self):
        User = private.create_object_factory(
            id='property',
            friend='link',
        )

        u1 = User(1, None)
        u2 = User(2, u1)
        u3 = User(3, edgedb.Set([]))

        self.assertEqual(set(dir(u2['friend'])), {'source', 'target'})

        self.assertIs(u2['friend'].target, u1)

        self.assertIsNone(u1['friend'])

        self.assertEqual(len(u3['friend']), 0)
        self.assertEqual(
            repr(u3['friend']),
            "LinkSet(name='friend', source_id=3, target_ids={})")

        self.assertEqual(
            repr(u2['friend']),
            "Link(name='friend', source_id=2, target_id=1)")

    def test_object_links_4(self):
        User = private.create_object_factory(
            id='property',
            friend='link',
        )

        u = User(1, None)

        with self.assertRaisesRegex(KeyError,
                                    "link 'error_key' does not exist"):
            u['error_key']


class TestSet(unittest.TestCase):

    def test_set_1(self):
        s = edgedb.Set(())
        self.assertEqual(repr(s), 'Set{}')

        s = edgedb.Set((1, 2, [], 'a'))

        self.assertEqual(s[1], 2)
        self.assertEqual(s[2], [])
        self.assertEqual(len(s), 4)
        with self.assertRaises(IndexError):
            s[10]

        with self.assertRaises(TypeError):
            s[0] = 1

    def test_set_2(self):
        s = edgedb.Set((1, 2, 3000, 'a'))

        self.assertEqual(repr(s), "Set{1, 2, 3000, 'a'}")

        self.assertEqual(
            hash(s),
            hash(edgedb.Set((1, 2, sum([1000, 2000]), 'a'))))

        self.assertNotEqual(
            hash(s),
            hash((1, 2, 3000, 'a')))

    def test_set_3(self):
        s = edgedb.Set(())

        self.assertEqual(len(s), 0)
        self.assertEqual(hash(s), hash(edgedb.Set(())))
        self.assertNotEqual(hash(s), hash(()))

    def test_set_4(self):
        s = edgedb.Set(([],))
        s[0].append(s)
        self.assertEqual(repr(s), "Set{[Set{...}]}")

    def test_set_5(self):
        self.assertEqual(
            edgedb.Set([1, 2, 3]),
            edgedb.Set([3, 2, 1]))

        self.assertEqual(
            edgedb.Set([]),
            edgedb.Set([]))

        self.assertEqual(
            edgedb.Set([1]),
            edgedb.Set([1]))

        self.assertNotEqual(
            edgedb.Set([1]),
            edgedb.Set([]))

        self.assertNotEqual(
            edgedb.Set([1]),
            edgedb.Set([2]))

        self.assertNotEqual(
            edgedb.Set([1, 2]),
            edgedb.Set([2, 2]))

        self.assertNotEqual(
            edgedb.Set([1, 1, 2]),
            edgedb.Set([2, 2, 1]))

    def test_set_6(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o1 = f(1, 'aa', edgedb.Set([1, 2, 3]))
        o2 = f(1, 'ab', edgedb.Set([1, 2, 4]))
        o3 = f(3, 'ac', edgedb.Set([5, 5, 5, 5]))

        self.assertEqual(
            edgedb.Set([o1, o2, o3]),
            edgedb.Set([o2, o3, o1]))

        self.assertEqual(
            edgedb.Set([o1, o3]),
            edgedb.Set([o2, o3]))

        self.assertNotEqual(
            edgedb.Set([o1, o1]),
            edgedb.Set([o2, o3]))

    def test_set_7(self):
        self.assertEqual(
            edgedb.Set([1, 2, 3]),
            [1, 2, 3])

        self.assertNotEqual(
            edgedb.Set([1, 2, 3]),
            [3, 2, 1])

        self.assertNotEqual(
            edgedb.Set([1, 2, 3]),
            1)

    def test_set_8(self):
        s = edgedb.Set([1, 2, 3])
        si = iter(s)
        self.assertEqual(list(si), [1, 2, 3])


class TestArray(unittest.TestCase):

    def test_array_empty_1(self):
        t = edgedb.Array()
        self.assertEqual(len(t), 0)
        self.assertNotEqual(hash(t), hash(()))
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[0]
        self.assertEqual(repr(t), "[]")

    def test_array_2(self):
        t = edgedb.Array((1, 'a'))

        self.assertEqual(repr(t), "[1, 'a']")
        self.assertEqual(str(t), "[1, 'a']")

        self.assertEqual(len(t), 2)
        self.assertEqual(hash(t), hash(edgedb.Array([1, 'a'])))
        self.assertNotEqual(hash(t), hash(edgedb.Array([10, 'ab'])))

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

    def test_array_3(self):
        t = edgedb.Array((1, []))
        t[1].append(t)
        self.assertEqual(t[1], [t])
        self.assertEqual(repr(t), '[1, [[...]]]')

    def test_array_4(self):
        self.assertEqual(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 3, 2]))

        self.assertLess(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 3, 2]))

        self.assertEqual(
            edgedb.Array([]),
            edgedb.Array([]))

        self.assertEqual(
            edgedb.Array([1]),
            edgedb.Array([1]))

        self.assertGreaterEqual(
            edgedb.Array([1]),
            edgedb.Array([1]))

        self.assertNotEqual(
            edgedb.Array([1]),
            edgedb.Array([]))

        self.assertGreater(
            edgedb.Array([1]),
            edgedb.Array([]))

        self.assertNotEqual(
            edgedb.Array([1]),
            edgedb.Array([2]))

        self.assertLess(
            edgedb.Array([1]),
            edgedb.Array([2]))

        self.assertNotEqual(
            edgedb.Array([1, 2]),
            edgedb.Array([2, 2]))

        self.assertNotEqual(
            edgedb.Array([1, 1]),
            edgedb.Array([2, 2, 1]))

    def test_array_5(self):
        self.assertEqual(
            edgedb.Array([1, 2, 3]),
            [1, 2, 3])

        self.assertEqual(
            [1, 2, 3],
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            [1, 2, 4],
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            [1, 3, 2])

        self.assertLess(
            edgedb.Array([1, 2, 3]),
            [1, 3, 2])

        self.assertEqual(
            edgedb.Array([]),
            [])

        self.assertEqual(
            edgedb.Array([1]),
            [1])

        self.assertGreaterEqual(
            edgedb.Array([1]),
            [1])

        self.assertNotEqual(
            edgedb.Array([1]),
            [])

        self.assertGreater(
            edgedb.Array([1]),
            [])

        self.assertNotEqual(
            edgedb.Array([1]),
            [2])

        self.assertLess(
            edgedb.Array([1]),
            [2])

        self.assertNotEqual(
            edgedb.Array([1, 2]),
            [2, 2])

        self.assertNotEqual(
            edgedb.Array([1, 1]),
            [2, 2, 1])

    def test_array_6(self):
        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            False)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from libc.stdint cimport int64_t, int32_t


DEF MONTHS_PER_YEAR         = 12
DEF USECS_PER_HOUR          = 3600000000
DEF USECS_PER_MINUTE        = 60000000
DEF USECS_PER_SEC           = 1000000
DEF MAX_INTERVAL_PRECISION  = 6


@cython.final
cdef class RelativeDuration:

    def __init__(self, *, int64_t microseconds=0,
                 int32_t days=0, int32_t months=0):
        self.microseconds = microseconds
        self.days = days
        self.months = months

    def __eq__(self, other):
        if type(other) is not RelativeDuration:
            return NotImplemented

        return (
            self.microseconds == (<RelativeDuration>other).microseconds and
            self.days == (<RelativeDuration>other).days and
            self.months == (<RelativeDuration>other).months
        )

    def __hash__(self):
        return hash((RelativeDuration, self.microseconds, self.days, self.months))

    def __repr__(self):
        return f'<edgedb.RelativeDuration "{self}">'

    @cython.cdivision(True)
    def __str__(self):
        # This is closely modeled after interval_out().
        cdef:
            int64_t year = self.months / MONTHS_PER_YEAR
            int64_t mon = self.months % MONTHS_PER_YEAR
            int64_t time = self.microseconds
            int64_t tfrac
            int64_t hour
            int64_t min
            int64_t sec
            int64_t fsec
            list buf = []
            bint neg
            bint is_before = False
            bint is_first = True

        if not self.months and not self.days and not time:
            return 'PT0S'

        tfrac = time / <int64_t>USECS_PER_HOUR
        time -= tfrac * <int64_t>USECS_PER_HOUR
        hour = tfrac

        if (hour < 0) != (tfrac < 0):
            raise ValueError('interval out of range')

        tfrac = time / <int64_t>USECS_PER_MINUTE
        time -= tfrac * <int64_t>USECS_PER_MINUTE
        min = tfrac
        sec = time / USECS_PER_SEC
        fsec = time - sec * USECS_PER_SEC

        buf.append('P')

        if year:
            buf.append(f'{year}Y')

        if mon:
            buf.append(f'{mon}M')

        if self.days:
            buf.append(f'{self.days}D')

        if not self.microseconds:
            return ''.join(buf)

        buf.append('T')

        if hour:
            buf.append(f'{hour}H')

        if min:
            buf.append(f'{min}M')

        if sec or fsec:
            sign = '-' if min < 0 or fsec < 0 else ''
            buf.append(f'{sign}{abs(sec)}')

            if fsec:
                buf.append(f'.{abs(fsec):0>6}'.rstrip('0'))

            buf.append('S')

        return ''.join(buf)


cdef new_duration(int64_t microseconds, int32_t days, int32_t months):
    cdef RelativeDuration dur = RelativeDuration.__new__(RelativeDuration)

    dur.microseconds = microseconds
    dur.days = days
    dur.months = months

    return dur



import abc
import enum
import random
from collections import namedtuple

from . import errors


_RetryRule = namedtuple("_RetryRule", ["attempts", "backoff"])


def default_backoff(attempt):
    return (2 ** attempt) * 0.1 + random.randrange(100) * 0.001


class RetryCondition:
    """Specific condition to retry on for fine-grained control"""
    TransactionConflict = enum.auto()
    NetworkError = enum.auto()


class IsolationLevel:
    """Isolation level for transaction"""
    Serializable = "SERIALIZABLE"


class RetryOptions:
    """An immutable class that contains rules for `transaction()`"""
    __slots__ = ['_default', '_overrides']

    def __init__(self, attempts: int, backoff=default_backoff):
        self._default = _RetryRule(attempts, backoff)
        self._overrides = None

    def with_rule(self, condition, attempts=None, backoff=None):
        default = self._default
        overrides = self._overrides
        if overrides is None:
            overrides = {}
        else:
            overrides = overrides.copy()
        overrides[condition] = _RetryRule(
            default.attempts if attempts is None else attempts,
            default.backoff if backoff is None else backoff,
        )
        result = RetryOptions.__new__(RetryOptions)
        result._default = default
        result._overrides = overrides
        return result

    @classmethod
    def defaults(cls):
        return cls(
            attempts=3,
            backoff=default_backoff,
        )

    def get_rule_for_exception(self, exception):
        default = self._default
        overrides = self._overrides
        res = default
        if overrides:
            if isinstance(exception, errors.TransactionConflictError):
                res = overrides.get(RetryCondition.TransactionConflict, res)
            elif isinstance(exception, errors.ClientError):
                res = overrides.get(RetryCondition.NetworkError, res)
        return res


class TransactionOptions:
    """Options for `transaction()`"""
    __slots__ = ['_isolation', '_readonly', '_deferrable']

    def __init__(
        self,
        isolation: IsolationLevel=IsolationLevel.Serializable,
        readonly: bool = False,
        deferrable: bool = False,
    ):
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable

    @classmethod
    def defaults(cls):
        return cls()

    def start_transaction_query(self):
        isolation = str(self._isolation)
        if self._readonly:
            mode = 'READ ONLY'
        else:
            mode = 'READ WRITE'

        if self._deferrable:
            defer = 'DEFERRABLE'
        else:
            defer = 'NOT DEFERRABLE'

        return f'START TRANSACTION ISOLATION {isolation}, {mode}, {defer};'

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'isolation:{self._isolation}, '
            f'readonly:{self._readonly}, '
            f'deferrable:{self._deferrable}>'
        )


class _OptionsMixin:
    def __init__(self, *args, **kwargs):
        self._options = _Options.defaults()
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def _shallow_clone(self):
        pass

    def with_transaction_options(self, options: TransactionOptions = None):
        """Returns object with adjusted options for future transactions.

        :param options TransactionOptions:
            Object that encapsulates transaction options.

        This method returns a "shallow copy" of the current object
        with modified transaction options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.

        Transaction options are are used by the ``transaction`` method.
        """
        result = self._shallow_clone()
        result._options = self._options.with_transaction_options(options)
        return result

    def with_retry_options(self, options: RetryOptions=None):
        """Returns object with adjusted options for future retrying
        transactions.

        :param options RetryOptions:
            Object that encapsulates retry options.

        This method returns a "shallow copy" of the current object
        with modified transaction options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.
        """

        result = self._shallow_clone()
        result._options = self._options.with_retry_options(options)
        return result


class _Options:
    """Internal class for storing connection options"""

    __slots__ = ['_retry_options', '_transaction_options']

    def __init__(
        self,
        retry_options: RetryOptions,
        transaction_options: TransactionOptions,
    ):
        self._retry_options = retry_options
        self._transaction_options = transaction_options

    @property
    def retry_options(self):
        return self._retry_options

    @property
    def transaction_options(self):
        return self._transaction_options

    def with_retry_options(self, options: RetryOptions):
        return _Options(
            options,
            self._transaction_options,
        )

    def with_transaction_options(self, options: TransactionOptions):
        return _Options(
            self._retry_options,
            options,
        )

    @classmethod
    def defaults(cls):
        return cls(
            RetryOptions.defaults(),
            TransactionOptions.defaults(),
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import errno
import os
import re
import ssl
import typing
import urllib.parse
import warnings
import hashlib

from . import errors
from . import credentials
from . import platform


EDGEDB_PORT = 5656
ERRNO_RE = re.compile(r"\[Errno (\d+)\]")
TEMPORARY_ERRORS = (
    ConnectionAbortedError,
    ConnectionRefusedError,
    ConnectionResetError,
    FileNotFoundError,
)
TEMPORARY_ERROR_CODES = frozenset({
    errno.ECONNREFUSED,
    errno.ECONNABORTED,
    errno.ECONNRESET,
    errno.ENOENT,
})


class ClientConfiguration(typing.NamedTuple):

    connect_timeout: float
    command_timeout: float
    wait_until_available: float


def _validate_port_spec(hosts, port):
    if isinstance(port, list):
        # If there is a list of ports, its length must
        # match that of the host list.
        if len(port) != len(hosts):
            raise errors.InterfaceError(
                'could not match {} port numbers to {} hosts'.format(
                    len(port), len(hosts)))
    else:
        port = [port for _ in range(len(hosts))]

    return port


def _parse_hostlist(hostlist, port):
    if ',' in hostlist:
        # A comma-separated list of host addresses.
        hostspecs = hostlist.split(',')
    else:
        hostspecs = [hostlist]

    hosts = []
    hostlist_ports = []

    if not port:
        portspec = os.environ.get('EDGEDB_PORT')
        if portspec:
            if ',' in portspec:
                default_port = [int(p) for p in portspec.split(',')]
            else:
                default_port = int(portspec)
        else:
            default_port = EDGEDB_PORT

        default_port = _validate_port_spec(hostspecs, default_port)

    else:
        port = _validate_port_spec(hostspecs, port)

    for i, hostspec in enumerate(hostspecs):
        addr, _, hostspec_port = hostspec.partition(':')
        hosts.append(addr)

        if not port:
            if hostspec_port:
                hostlist_ports.append(int(hostspec_port))
            else:
                hostlist_ports.append(default_port[i])

    if not port:
        port = hostlist_ports

    return hosts, port


def _hash_path(path):
    path = os.path.realpath(path)
    if platform.IS_WINDOWS and not path.startswith('\\\\'):
        path = '\\\\?\\' + path
    return hashlib.sha1(str(path).encode('utf-8')).hexdigest()


def _stash_path(path):
    base_name = os.path.basename(path)
    dir_name = base_name + '-' + _hash_path(path)
    return platform.search_config_dir('projects', dir_name)


def _validate_tls_security(val: str) -> str:
    val = val.lower()
    if val not in {"insecure", "no_host_verification", "strict", "default"}:
        raise ValueError(
            "tls_security can only be one of "
            "`insecure`, `no_host_verification`, `strict` or `default`"
        )

    return val


class ResolvedConnectConfig:
    _host = None
    _host_source = None

    _port = None
    _port_source = None

    _database = None
    _database_source = None

    _user = None
    _user_source = None

    _password = None
    _password_source = None

    _tls_ca_data = None
    _tls_ca_data_source = None

    _tls_security = None
    _tls_security_source = None

    server_settings = {}

    def _set_param(self, param, value, source, validator=None):
        param_name = '_' + param
        if getattr(self, param_name) is None:
            setattr(self, param_name + '_source', source)
            if value is not None:
                setattr(
                    self,
                    param_name,
                    validator(value) if validator else value
                )

    def set_host(self, host, source):
        self._set_param('host', host, source, _validate_host)

    def set_port(self, port, source):
        self._set_param('port', port, source, _validate_port)

    def set_database(self, database, source):
        self._set_param('database', database, source, _validate_database)

    def set_user(self, user, source):
        self._set_param('user', user, source, _validate_user)

    def set_password(self, password, source):
        self._set_param('password', password, source)

    def set_tls_ca_data(self, ca_data, source):
        self._set_param('tls_ca_data', ca_data, source)

    def set_tls_ca_file(self, ca_file, source):
        def read_ca_file(file_path):
            with open(file_path) as f:
                return f.read()

        self._set_param('tls_ca_data', ca_file, source, read_ca_file)

    def set_tls_security(self, security, source):
        self._set_param('tls_security', security, source,
                        _validate_tls_security)

    def add_server_settings(self, server_settings):
        _validate_server_settings(server_settings)
        self.server_settings = {**server_settings, **self.server_settings}

    @property
    def address(self):
        return (
            self._host if self._host else 'localhost',
            self._port if self._port else 5656
        )

    @property
    def database(self):
        return self._database if self._database else 'edgedb'

    @property
    def user(self):
        return self._user if self._user else 'edgedb'

    @property
    def password(self):
        return self._password

    @property
    def tls_security(self):
        tls_security = self._tls_security or 'default'
        security = os.environ.get('EDGEDB_CLIENT_SECURITY') or 'default'
        if security not in {'default', 'insecure_dev_mode', 'strict'}:
            raise ValueError(
                f'environment variable EDGEDB_CLIENT_SECURITY should be '
                f'one of strict, insecure_dev_mode or default, '
                f'got: {security!r}')

        if security == 'default':
            pass
        elif security == 'insecure_dev_mode':
            if tls_security == 'default':
                tls_security = 'insecure'
        elif security == 'strict':
            if tls_security == 'default':
                tls_security = 'strict'
            elif tls_security in {'no_host_verification', 'insecure'}:
                raise ValueError(
                    f'EDGEDB_CLIENT_SECURITY=strict but '
                    f'tls_security={tls_security}, tls_security must be '
                    f'set to strict when EDGEDB_CLIENT_SECURITY is strict')

        if tls_security != 'default':
            return tls_security

        if self._tls_ca_data is not None:
            return "no_host_verification"

        return "strict"

    _ssl_ctx = None

    @property
    def ssl_ctx(self):
        if (self._ssl_ctx):
            return self._ssl_ctx

        self._ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

        if self._tls_ca_data:
            self._ssl_ctx.load_verify_locations(
                cadata=self._tls_ca_data
            )
        else:
            self._ssl_ctx.load_default_certs(ssl.Purpose.SERVER_AUTH)
            if platform.IS_WINDOWS:
                import certifi
                self._ssl_ctx.load_verify_locations(cafile=certifi.where())

        tls_security = self.tls_security
        self._ssl_ctx.check_hostname = tls_security == "strict"

        if tls_security in {"strict", "no_host_verification"}:
            self._ssl_ctx.verify_mode = ssl.CERT_REQUIRED
        else:
            self._ssl_ctx.verify_mode = ssl.CERT_NONE

        self._ssl_ctx.set_alpn_protocols(['edgedb-binary'])

        return self._ssl_ctx


def _validate_host(host):
    if '/' in host:
        raise ValueError('unix socket paths not supported')
    if host == '' or ',' in host:
        raise ValueError(f'invalid host: "{host}"')
    return host


def _prepare_host_for_dsn(host):
    host = _validate_host(host)
    if ':' in host:
        # IPv6
        host = f'[{host}]'
    return host


def _validate_port(port):
    try:
        if isinstance(port, str):
            port = int(port)
        if not isinstance(port, int):
            raise ValueError()
    except Exception:
        raise ValueError(f'invalid port: {port}, not an integer')
    if port < 1 or port > 65535:
        raise ValueError(f'invalid port: {port}, must be between 1 and 65535')
    return port


def _validate_database(database):
    if database == '':
        raise ValueError(f'invalid database name: {database}')
    return database


def _validate_user(user):
    if user == '':
        raise ValueError(f'invalid user name: {user}')
    return user


def _validate_server_settings(server_settings):
    if (
        not isinstance(server_settings, dict) or
        not all(isinstance(k, str) for k in server_settings) or
        not all(isinstance(v, str) for v in server_settings.values())
    ):
        raise ValueError(
            'server_settings is expected to be None or '
            'a Dict[str, str]')


def _parse_connect_dsn_and_args(*, dsn, credentials_file, host, port, user,
                                password, database,
                                tls_ca_file, tls_security,
                                server_settings):

    resolved_config = ResolvedConnectConfig()

    dsn, instance_name = (
        (dsn, None)
        if dsn is not None and re.match('(?i)^[a-z]+://', dsn)
        else (None, dsn)
    )

    has_compound_options = _resolve_config_options(
        resolved_config,
        'Cannot have more than one of the following connection options: '
        + '"dsn", "credentials_file" or "host"/"port"',
        dsn=(dsn, '"dsn" option') if dsn is not None else None,
        instance_name=(
            (instance_name, '"dsn" option (parsed as instance name)')
            if instance_name is not None else None
        ),
        credentials_file=(
            (credentials_file, '"credentials_file" option')
            if credentials_file is not None else None
        ),
        host=(host, '"host" option') if host is not None else None,
        port=(port, '"port" option') if port is not None else None,
        database=(
            (database, '"database" option')
            if database is not None else None
        ),
        user=(user, '"user" option') if user is not None else None,
        password=(
            (password, '"password" option')
            if password is not None else None
        ),
        tls_ca_file=(
            (tls_ca_file, '"tls_ca_file" option')
            if tls_ca_file is not None else None
        ),
        tls_security=(
            (tls_security, '"tls_security" option')
            if tls_security is not None else None
        ),
        server_settings=(
            (server_settings, '"server_settings" option')
            if server_settings is not None else None
        ),
    )

    if has_compound_options is False:
        env_port = os.getenv("EDGEDB_PORT")
        if (
            resolved_config._port is None and
            env_port and env_port.startswith('tcp://')
        ):
            # EDGEDB_PORT is set by 'docker --link' so ignore and warn
            warnings.warn('EDGEDB_PORT in "tcp://host:port" format, ' +
                          'so will be ignored')
            env_port = None

        env_dsn = os.getenv('EDGEDB_DSN')
        env_instance = os.getenv('EDGEDB_INSTANCE')
        env_credentials_file = os.getenv('EDGEDB_CREDENTIALS_FILE')
        env_host = os.getenv('EDGEDB_HOST')
        env_database = os.getenv('EDGEDB_DATABASE')
        env_user = os.getenv('EDGEDB_USER')
        env_password = os.getenv('EDGEDB_PASSWORD')
        env_tls_ca_file = os.getenv('EDGEDB_TLS_CA_FILE')
        env_tls_security = os.getenv('EDGEDB_CLIENT_TLS_SECURITY')

        has_compound_options = _resolve_config_options(
            resolved_config,
            'Cannot have more than one of the following connection '
            + 'environment variables: "EDGEDB_DSN", "EDGEDB_INSTANCE", '
            + '"EDGEDB_CREDENTIALS_FILE" or "EDGEDB_HOST"/"EDGEDB_PORT"',
            dsn=(
                (env_dsn, '"EDGEDB_DSN" environment variable')
                if env_dsn is not None else None
            ),
            instance_name=(
                (env_instance, '"EDGEDB_INSTANCE" environment variable')
                if env_instance is not None else None
            ),
            credentials_file=(
                (env_credentials_file,
                 '"EDGEDB_CREDENTIALS_FILE" environment variable')
                if env_credentials_file is not None else None
            ),
            host=(
                (env_host, '"EDGEDB_HOST" environment variable')
                if env_host is not None else None
            ),
            port=(
                (env_port, '"EDGEDB_PORT" environment variable')
                if env_port is not None else None
            ),
            database=(
                (env_database, '"EDGEDB_DATABASE" environment variable')
                if env_database is not None else None
            ),
            user=(
                (env_user, '"EDGEDB_USER" environment variable')
                if env_user is not None else None
            ),
            password=(
                (env_password, '"EDGEDB_PASSWORD" environment variable')
                if env_password is not None else None
            ),
            tls_ca_file=(
                (env_tls_ca_file, '"EDGEDB_TLS_CA_FILE" environment variable')
                if env_tls_ca_file is not None else None
            ),
            tls_security=(
                (env_tls_security,
                 '"EDGEDB_CLIENT_TLS_SECURITY" environment variable')
                if env_tls_security is not None else None
            ),
        )

    if has_compound_options is False:
        dir = find_edgedb_project_dir()
        stash_dir = _stash_path(dir)
        if os.path.exists(stash_dir):
            with open(os.path.join(stash_dir, 'instance-name'), 'rt') as f:
                instance_name = f.read().strip()

                _resolve_config_options(
                    resolved_config,
                    '',
                    instance_name=(
                        instance_name,
                        f'project linked instance ("{instance_name}")'
                    )
                )
        else:
            raise errors.ClientConnectionError(
                f'Found `edgedb.toml` but the project is not initialized. '
                f'Run `edgedb project init`.'
            )

    return resolved_config


def _parse_dsn_into_config(
    resolved_config: ResolvedConnectConfig,
    dsn: typing.Tuple[str, str]
):
    dsn_str, source = dsn

    try:
        parsed = urllib.parse.urlparse(dsn_str)
        host = (
            urllib.parse.unquote(parsed.hostname) if parsed.hostname else None
        )
        port = parsed.port
        database = parsed.path
        user = parsed.username
        password = parsed.password
    except Exception as e:
        raise ValueError(f'invalid DSN or instance name: {str(e)}')

    if parsed.scheme != 'edgedb':
        raise ValueError(
            f'invalid DSN: scheme is expected to be '
            f'"edgedb", got {parsed.scheme!r}')

    query = (
        urllib.parse.parse_qs(parsed.query, keep_blank_values=True)
        if parsed.query != ''
        else {}
    )
    for key, val in query.items():
        if isinstance(val, list):
            if len(val) > 1:
                raise ValueError(
                    f'invalid DSN: duplicate query parameter {key}')
            query[key] = val[-1]

    def handle_dsn_part(
        paramName, value, currentValue, setter,
        formatter=lambda val: val
    ):
        param_values = [
            (value if value != '' else None),
            query.get(paramName),
            query.get(paramName + '_env'),
            query.get(paramName + '_file')
        ]
        if len([p for p in param_values if p is not None]) > 1:
            raise ValueError(
                f'invalid DSN: more than one of ' +
                f'{(paramName + ", ") if value else ""}' +
                f'?{paramName}=, ?{paramName}_env=, ?{paramName}_file= ' +
                f'was specified'
            )

        if currentValue is None:
            param = (
                value if (value is not None and value != '')
                else query.get(paramName)
            )
            paramSource = source

            if param is None:
                env = query.get(paramName + '_env')
                if env is not None:
                    param = os.getenv(env)
                    if param is None:
                        raise ValueError(
                            f'{paramName}_env environment variable "{env}" ' +
                            f'doesn\'t exist')
                    paramSource = paramSource + f' ({paramName}_env: {env})'
            if param is None:
                filename = query.get(paramName + '_file')
                if filename is not None:
                    with open(filename) as f:
                        param = f.read()
                    paramSource = (
                        paramSource + f' ({paramName}_file: {filename})'
                    )

            param = formatter(param) if param is not None else None

            setter(param, paramSource)

        query.pop(paramName, None)
        query.pop(paramName + '_env', None)
        query.pop(paramName + '_file', None)

    handle_dsn_part(
        'host', host, resolved_config._host, resolved_config.set_host
    )

    handle_dsn_part(
        'port', port, resolved_config._port, resolved_config.set_port
    )

    def strip_leading_slash(str):
        return str[1:] if str.startswith('/') else str

    handle_dsn_part(
        'database', strip_leading_slash(database),
        resolved_config._database, resolved_config.set_database,
        strip_leading_slash
    )

    handle_dsn_part(
        'user', user, resolved_config._user, resolved_config.set_user
    )

    handle_dsn_part(
        'password', password,
        resolved_config._password, resolved_config.set_password
    )

    handle_dsn_part(
        'tls_cert_file', None,
        resolved_config._tls_ca_data, resolved_config.set_tls_ca_file
    )

    handle_dsn_part(
        'tls_security', None,
        resolved_config._tls_security,
        resolved_config.set_tls_security
    )

    resolved_config.add_server_settings(query)


def _resolve_config_options(
    resolved_config: ResolvedConnectConfig,
    compound_error: str,
    *,
    dsn=None,
    instance_name=None,
    credentials_file=None,
    host=None,
    port=None,
    database=None,
    user=None,
    password=None,
    tls_ca_file=None,
    tls_security=None,
    server_settings=None
):
    if database is not None:
        resolved_config.set_database(*database)
    if user is not None:
        resolved_config.set_user(*user)
    if password is not None:
        resolved_config.set_password(*password)
    if tls_ca_file is not None:
        resolved_config.set_tls_ca_file(*tls_ca_file)
    if tls_security is not None:
        resolved_config.set_tls_security(*tls_security)
    if server_settings is not None:
        resolved_config.add_server_settings(server_settings[0])

    compound_params = [dsn, instance_name, credentials_file, host or port]
    compound_params_count = len([p for p in compound_params if p is not None])

    if compound_params_count > 1:
        raise errors.ClientConnectionError(compound_error)

    if compound_params_count == 1:
        if dsn is not None or host is not None or port is not None:
            if port is not None:
                resolved_config.set_port(*port)
            if dsn is None:
                dsn = (
                    'edgedb://' +
                    (_prepare_host_for_dsn(host[0]) if host else ''),
                    host[1] if host is not None else port[1]
                )
            _parse_dsn_into_config(resolved_config, dsn)
        else:
            if credentials_file is None:
                if (
                    re.match(
                        '^[A-Za-z_][A-Za-z_0-9]*$',
                        instance_name[0]
                    ) is None
                ):
                    raise ValueError(
                        f'invalid DSN or instance name: "{instance_name[0]}"'
                    )
                credentials_file = (
                    credentials.get_credentials_path(instance_name[0]),
                    instance_name[1]
                )
            creds = credentials.read_credentials(credentials_file[0])

            source = credentials_file[1]

            resolved_config.set_host(creds.get('host'), source)
            resolved_config.set_port(creds.get('port'), source)
            resolved_config.set_database(creds.get('database'), source)
            resolved_config.set_user(creds.get('user'), source)
            resolved_config.set_password(creds.get('password'), source)
            resolved_config.set_tls_ca_data(creds.get('tls_cert_data'), source)
            resolved_config.set_tls_security(
                creds.get('tls_security'),
                source
            )

        return True

    return False


def find_edgedb_project_dir():
    dir = os.getcwd()
    dev = os.stat(dir).st_dev

    while True:
        toml = os.path.join(dir, 'edgedb.toml')
        if not os.path.isfile(toml):
            parent = os.path.dirname(dir)
            if parent == dir:
                raise errors.ClientConnectionError(
                    f'no `edgedb.toml` found and '
                    f'no connection options specified'
                )
            parent_dev = os.stat(parent).st_dev
            if parent_dev != dev:
                raise errors.ClientConnectionError(
                    f'no `edgedb.toml` found and '
                    f'no connection options specified'
                    f'(stopped searching for `edgedb.toml` at file system'
                    f'boundary {dir!r})'
                )
            dir = parent
            dev = parent_dev
            continue
        return dir


def parse_connect_arguments(
    *, dsn, credentials_file, host, port,
    database, user, password,
    tls_ca_file, tls_security,
    timeout, command_timeout, wait_until_available,
    server_settings
) -> typing.Tuple[ResolvedConnectConfig, ClientConfiguration]:

    if command_timeout is not None:
        try:
            if isinstance(command_timeout, bool):
                raise ValueError
            command_timeout = float(command_timeout)
            if command_timeout <= 0:
                raise ValueError
        except ValueError:
            raise ValueError(
                'invalid command_timeout value: '
                'expected greater than 0 float (got {!r})'.format(
                    command_timeout)) from None

    connect_config = _parse_connect_dsn_and_args(
        dsn=dsn, credentials_file=credentials_file, host=host, port=port,
        database=database, user=user, password=password,
        tls_ca_file=tls_ca_file, tls_security=tls_security,
        server_settings=server_settings,
    )

    client_config = ClientConfiguration(
        connect_timeout=timeout,
        command_timeout=command_timeout,
        wait_until_available=wait_until_available or 0,
    )

    return connect_config, client_config


def check_alpn_protocol(ssl_obj):
    if ssl_obj.selected_alpn_protocol() != 'edgedb-binary':
        raise errors.ClientConnectionFailedError(
            "The server doesn't support the edgedb-binary protocol."
        )


def render_client_no_connection_error(prefix, addr, attempts, duration):
    if isinstance(addr, str):
        msg = (
            f'{prefix}'
            f'\n\tAfter {attempts} attempts in {duration:.1f} sec'
            f'\n\tIs the server running locally and accepting '
            f'\n\tconnections on Unix domain socket {addr!r}?'
        )
    else:
        msg = (
            f'{prefix}'
            f'\n\tAfter {attempts} attempts in {duration:.1f} sec'
            f'\n\tIs the server running on host {addr[0]!r} '
            f'and accepting '
            f'\n\tTCP/IP connections on port {addr[1]}?'
        )
    return msg


def _extract_errno(s):
    """Extract multiple errnos from error string

    When we connect to a host that has multiple underlying IP addresses, say
    ``localhost`` having ``::1`` and ``127.0.0.1``, we get
    ``OSError("Multiple exceptions:...")`` error without ``.errno`` attribute
    set. There are multiple ones in the text, so we extract all of them.
    """
    result = []
    for match in ERRNO_RE.finditer(s):
        result.append(int(match.group(1)))
    if result:
        return result


def wrap_error(e):
    message = str(e)
    if e.errno is None:
        errnos = _extract_errno(message)
    else:
        errnos = [e.errno]

    if errnos:
        is_temp = any((code in TEMPORARY_ERROR_CODES for code in errnos))
    else:
        is_temp = isinstance(e, TEMPORARY_ERRORS)

    if is_temp:
        return errors.ClientConnectionFailedTemporarilyError(message)
    else:
        return errors.ClientConnectionFailedError(message)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


@cython.final
cdef class ScalarCodec(BaseCodec):

    def __cinit__(self):
        self.c_encoder = NULL
        self.c_decoder = NULL

    cdef encode(self, WriteBuffer buf, object obj):
        self.c_encoder(DEFAULT_CODEC_CONTEXT, buf, obj)

    cdef decode(self, FRBuffer *buf):
        return self.c_decoder(DEFAULT_CODEC_CONTEXT, buf)

    cdef derive(self, bytes tid):
        return ScalarCodec.new(tid, self.name, self.c_encoder, self.c_decoder)

    @staticmethod
    cdef BaseCodec new(bytes tid, str name,
                       pgproto.encode_func encoder,
                       pgproto.decode_func decoder):
        cdef:
            ScalarCodec codec

        codec = ScalarCodec.__new__(ScalarCodec)

        codec.tid = tid
        codec.name = name
        codec.c_encoder = encoder
        codec.c_decoder = decoder

        return codec



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import uuid

import edgedb
from edgedb import _testbase as tb


class TestEnum(tb.AsyncQueryTestCase):

    ISOLATED_METHODS = False

    SETUP = '''
        CREATE SCALAR TYPE CellType EXTENDING enum<'red', 'white'>;
        CREATE SCALAR TYPE Color EXTENDING enum<'red', 'white'>;
    '''

    async def test_enum_01(self):
        ct_red = await self.con.query_single('SELECT <CellType>"red"')
        ct_white = await self.con.query_single('SELECT <CellType>"white"')
        c_red = await self.con.query_single('SELECT <Color>"red"')

        self.assertTrue(isinstance(ct_red, edgedb.EnumValue))
        self.assertTrue(isinstance(ct_red.__tid__, uuid.UUID))

        self.assertEqual(repr(ct_red), "<EnumValue 'red'>")

        self.assertEqual(str(ct_red), 'red')
        self.assertNotEqual(ct_red, 'red')
        self.assertFalse(ct_red == 'red')
        self.assertFalse(ct_red == c_red)

        self.assertEqual(ct_red, ct_red)
        self.assertNotEqual(ct_red, ct_white)
        self.assertNotEqual(ct_red, c_red)

        self.assertLess(ct_red, ct_white)
        self.assertLessEqual(ct_red, ct_red)
        self.assertGreater(ct_white, ct_red)
        self.assertGreaterEqual(ct_white, ct_white)

        with self.assertRaises(TypeError):
            _ = ct_red < 'red'
        with self.assertRaises(TypeError):
            _ = ct_red > 'red'
        with self.assertRaises(TypeError):
            _ = ct_red <= 'red'
        with self.assertRaises(TypeError):
            _ = ct_red >= 'red'

        with self.assertRaises(TypeError):
            _ = ct_red < c_red
        with self.assertRaises(TypeError):
            _ = ct_red > c_red
        with self.assertRaises(TypeError):
            _ = ct_red <= c_red
        with self.assertRaises(TypeError):
            _ = ct_red >= c_red

        self.assertNotEqual(hash(ct_red), hash(c_red))
        self.assertNotEqual(hash(ct_red), hash('red'))
